﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2023 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "ARMChip.h"
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichtenausstatter  ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupPushMessage(Cy4Message* message)
{
    PushMessage pushMessage;
    STORING_Load(STO_PressedActionConfiguration.pushMessage, &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTextFeedMessage(Cy4Message* message)
{
    TextFeed textFeed;
    STORING_Load(STO_PressedActionConfiguration.textFeed, &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfragebearbeiter  ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: InfoLedSignals (read write)                                                 //
//                                                                                      //
//  Beschreibung: Bestimmt die Signalisierungen der LEDs. Folgende Lichtsignale  	    //
//                können eingestellt und abgefragt werden:                              //
//                                                                                      //
//                       0 = Kein Signal, Ausgeschaltet                                 //
//                       1 = Leuchtet durchgehend                                       //
//                       2 = Blinkt                                                     //
//                       3 = Herzschlag                                                 //
//                       4 = Bus Aktivität                                              //
//                       5 = Ein, wenn Taste gedrückt ist								//
//                       x = Wirkungslos                                                //
//                                                                                      //
//              0   1   2                                                               //
//            +---+---+---+                                                             //
//  Belegung: |SLD|RLD|BLD|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [SLD] Lichtsignal der System-LED, Werkseinstellung ist 3.                           //
//  [RLD] Lichtsignal der roten LED, Werkseinstellung ist 4.                            //
//  [BLD] Lichtsignal der blauen LED, Werkseinstellung ist 5.                           //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnInfoLedSignals(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[3];

    LedSignal ledSignal[3];
    STORING_Load(STO_LedSignal, ledSignal);
    for (UInt32 i=0; i<CARDINALITY(ledSignal); i++)
    {	data[i] = (UInt08)ledSignal[i];
    }
    CY4SERVICE_TransferPayload(service, message, data);
    for (UInt32 i=0; i<CARDINALITY(ledSignal); i++)
    {	if (data[i] < 6) ledSignal[i] = (LedSignal)data[i];
	}
	STORING_Save(ledSignal, STO_LedSignal);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: PressActionConfiguration (read write)                                       //
//                                                                                      //
//  Beschreibung: Bei einem Tastendruck kann eine Push-Nachricht und/oder eine Rund-    //
//                mitteilung verschickt werden.					                        //
//                                                                                      //
//              0                             7   8   9        56  57                   //
//            +---+  1   2   3   4   5   6  +---+---+---+-   -+---+---+                 //
//  Belegung: |ENA|+---+---+---+---+---+---+|T00|T01|T02| ... |T49| 0 |                 //
//            +---+|NET|DEV|LC0|LC1|LC2|PAR|+---+---+---+-   -+---+---+                 //
//                 +---+---+---+---+---+---+                                            //
//                                                                                      //
//  [ENA] Tastendruck-Aktion eingeschaltet: 0 = Aus, sonst Ein.							//
//  [NET] Push-Nachricht Ziel bei Tastendruck, Netzadresse   : 0 ..     255             //
//  [DEV] Push-Nachricht Ziel bei Tastendruck, Geräteadresse : 0 ..     255             //
//  [LCx] Push-Nachricht Ziel bei Tastendruck, Dienstort     : 0 .. 0xFFFFF             //
//  [PAR] Push-Nachricht Ziel bei Tastendruck, Parameter-Byte: 0 ..     255             //
//  [Txx] Zeichen der Rundmitteilung bei Tastendruck: 0 .. 255                          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnPressActionConfiguration(const Cy4Service* service, Cy4Message* message)
{
	PressedActionConfiguration configuration;
	STORING_Load(STO_PressedActionConfiguration, &configuration);
    UInt08  data[58];
    data[0] = configuration.enabled;
    PUSHMESSAGE_Serialize(&configuration.pushMessage, &data[1]);
    TEXTFEED_Serialize(&configuration.textFeed, &data[7]);
    CY4SERVICE_TransferPayload(service, message, data);
    configuration.enabled = (bool)data[0];
    PUSHMESSAGE_Deserialize(&data[1], &configuration.pushMessage);
    TEXTFEED_Deserialize(&data[7], &configuration.textFeed);
    STORING_Save(&configuration, STO_PressedActionConfiguration);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = typePlate[1][j];
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^').                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^').                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAddress(address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
    LedSignal ledSignal[3];
    STORING_Load(STO_LedSignal, ledSignal);
    UInt16 ticks = BOARD_MSecTicks();
    for (UInt32 i=0; i<CARDINALITY(ledSignal); i++)
    {	bool ledOn;
		switch (ledSignal[i])
		{   case LEDSIGNAL_OFF:          ledOn = false;                     break;
			case LEDSIGNAL_ON:           ledOn = true;                      break;
			case LEDSIGNAL_BLINKING:     ledOn = (bool)(ticks & 0x100);     break;
			case LEDSIGNAL_HEARTBEAT:    ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180); break;
			case LEDSIGNAL_BUS_ACTIVITY: ledOn = BOARD_BusActivity();       break;
			case LEDSIGNAL_PUSH_BUTTON:  ledOn = BOARD_PushButtonPressed(); break;
			default:                     ledOn = false;
		}
		Led led;
		switch (i)
		{	case 0:  led = LED_ONBOARD; break;
			case 1:  led = LED_RED;     break;
			case 2:  led = LED_BLUE;    break;
			default: return;
		}
		BOARD_SetLed(led, ledOn);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProccedPushButtonAction(void)
{
	static bool pressed = false;
	if (BOARD_PushButtonPressed())
	{	if (!pressed)
		{	pressed = true;
		    PressedActionConfiguration pressedActionConfiguration;
	    	STORING_Load(STO_PressedActionConfiguration, &pressedActionConfiguration);
	    	if (pressedActionConfiguration.enabled)
	    	{	if (PUSHMESSAGE_IsValid(&pressedActionConfiguration.pushMessage))
	    		{	if (!CY4NET_IsPending(SetupPushMessage, NULL))
	    			{	CY4NET_Post(SetupPushMessage, NULL);
	    			}
	    		}
	    		if (!TEXTFEED_IsEmpty(&pressedActionConfiguration.textFeed))
	    		{	if (!CY4NET_IsPending(SetupTextFeedMessage, NULL))
	    			{	CY4NET_Post(SetupTextFeedMessage, NULL);
	    			}
	    		}
	    	}
		}
	}
	else
	{	pressed = false;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                    service                    |          handler           |
        //+------------------+----+------------------...--+----------------------------+
    	{ CY4SERVICE(0x200000,  3 , CY4SERVICE_READ_WRITE), OnInfoLedSignals           },
   		{ CY4SERVICE(0x220000, 58 , CY4SERVICE_READ_WRITE), OnPressActionConfiguration },
        { CY4SERVICE(0xFFFF00, 61 , CY4SERVICE_READ_WRITE), OnDeviceLabel              },
        { CY4SERVICE(0xFFFF40,  1 , CY4SERVICE_WRITE_ONLY), OnFactorySettings          },
        { CY4SERVICE(0xFFFFF0,  7 , CY4SERVICE_READ_ONLY ), OnCallTypePlate            },
        { CY4SERVICE(0xFFFFF8,  1 , CY4SERVICE_WRITE_ONLY), OnEnterApplication         },
        { CY4SERVICE(0xFFFFFA,  1 , CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader       },
        { CY4SERVICE(0xFFFFFC,  2 , CY4SERVICE_WRITE_ONLY), OnChangeCy4Address         }
    };
    CY4NET_AssignServices(servicesTable);
    // Eigene Cy4-Addresse setzen
    {	Cy4Address address = STORING_LoadCy4SelfAddress();
    	CY4NET_SetAddress(address);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
	ProceedLedSignaling();
	ProccedPushButtonAction();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
