﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2023 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "stm32f3xx.h"
#include "stm32f3xx_ll_rcc.h"
#include "stm32f3xx_ll_bus.h"
#include "stm32f3xx_ll_system.h"
#include "stm32f3xx_ll_exti.h"
#include "stm32f3xx_ll_cortex.h"
#include "stm32f3xx_ll_utils.h"
#include "stm32f3xx_ll_pwr.h"
#include "stm32f3xx_ll_dma.h"
#include "stm32f3xx_ll_usart.h"
#include "stm32f3xx_ll_gpio.h"
#include "stm32f3xx_ll_wwdg.h"
#include "ARMChip.h"
#include "Board.h"
#include "Flash.h"
#include "Conf.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending;             	// · Übertragung im Gange
    bool   echo;             		// · Echo erhalten
    bool   error;             		// · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap;             	// · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei <byteGap> aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Konstanten  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

// ...

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt32      msecTicks                             = 0;
static volatile UInt32      secTicks                              = 0;
static volatile UInt08      busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08      busTxQueueHead                        = 0;
static volatile UInt08      busTxQueueLength                      = 0;
static volatile UInt08      busLoopbackByte                       = 0;
static volatile BusStatus   busStatus                             = { false, false, false, 0 };
static volatile BusRxItem   busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { false } };
static volatile UInt08      busRxQueueHead                        = 0;
static volatile UInt08      busRxQueueTail                        = 0;
static          UInt32      busActivityIndicator                  = 0;
static          DoNext      doNext                                = DO_STARTUP;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
	NVIC_SystemReset();		// Software-Reset: Neustart der Anwendung
}

//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
EnterPgmLdr(void)
{
	LL_APB1_GRP1_EnableClock(LL_APB1_GRP1_PERIPH_WWDG);		// Windows-Watchdog-Reset:
	LL_WWDG_SetCounter(WWDG, 63);							// Wechsel in den Programm-
	LL_WWDG_SetPrescaler(WWDG, LL_WWDG_PRESCALER_8);		// lader.
	LL_WWDG_Enable(WWDG);
	while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ResetWatchdog(void)
{
	IWDG->KR = 0x0000AAAA;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedIndicators(void)
{
    // 1ms Zeitbarriere
    UInt16 ticks = BOARD_MSecTicks();
    {	static UInt08 ms = 0;
        if (ms == (UInt08)ticks) return;
        ms = (UInt08)ticks;
    }
    // Bus-Aktivitätsindikator
    {   if (busActivityIndicator > 0)
        {   busActivityIndicator--;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
SysTick_Handler(void)
{
	msecTicks++;
	static int systemMillisecond = 0;
	if (systemMillisecond == 1000)
	{   systemMillisecond = 0;
		secTicks++;
	}
	// Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
        if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
USART1_IRQHandler(void)
{
	// IRQ Abfolge und Zusammenspiel der Signale <echo> und <pending>:
	//
	//          Start Tx:     |1|
	//                         :
	//          Tx-IRQ:  ···---:----------|2|----------|3|----···-|n|----------| |--···
	//          Rx-IRQ:  ···---:------|1|--:-------|2|--:-----···--:-------|n|--:---···
	//                      ___:       :___:        :___:          :        :_______
	//   Tx-IRQ Enabled: ...   |_______|   |________|   |_______...|________|       ...
	//                         :       :___:        :___:          :        :___:
	//          echo:    ...___:_______|   |________|   |_______...|________|   |___...
	//                         :________________________________________________:
	//          pending: ...___|                                                |___...
	//
	//  Bemerkung: Um sicherzustellen, dass die Ausführung der IRQs in der Reihenfolge
	//             Rx-Tx auftreten, wird der Tx-IRQ nach Absenden eines Datenbytes bis
	//             Eintritt des nächsten Rx-IRQs gesperrt.
	//

	if (LL_USART_IsActiveFlag_RXNE(USART1))
	{	// Rx
	    UInt08 byte = LL_USART_ReceiveData8(USART1);
	    LL_USART_EnableIT_TC(USART1);
	    busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
	    if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo ?
	    // Byte in Empfangsschlange stellen
	    UInt08 tail = busRxQueueTail;
	    busRxQueue[tail].data = byte;
	    tail++;
	    if (tail == CARDINALITY(busRxQueue)) tail = 0;
	    if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
	    else                        busRxQueueTail = tail;
	}
    if (LL_USART_IsActiveFlag_TC(USART1))
    {	// Tx
	    LL_USART_ClearFlag_TC(USART1);
		if (!busStatus.echo)
		{   busStatus.error = true;
			busTxQueueLength = 0;
		}
		busStatus.echo = false;
		UInt08 head = busTxQueueHead;
		if (busTxQueueLength == 0)
		{   busStatus.pending = false;
		}
		else
		{   busLoopbackByte = busTxQueue[head];
			LL_USART_DisableIT_TC(USART1);
			LL_USART_TransmitData8(USART1, busTxQueue[head]);
			head++;
			if (head == CARDINALITY(busTxQueue)) head = 0;
			busTxQueueLength--;
		}
		busTxQueueHead = head;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller      //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den eigentlichen   //
//  Reset auslöst.                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
	UInt16 ticks;
	Atomic
	{	ticks = (UInt16)msecTicks;
	}
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
	UInt16 ticks;
	Atomic
	{	ticks = (UInt16)secTicks;
	}
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als 'CFG_BUS_INTERBYTE_TIME-   //
//  OUT' Millisekunden eintritt.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in <byte> auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()      +--------------+ BOARD_BusTransmit()       //
//     |//////////+------------------------->|//////////////+--------------------+      //
//     |// IDLE //|                COMPLETED |// TRANSMIT //|                    |      //
//     |//////////|<-------------------------+//////////////|<-------------------+      //
//     +----------+                          +------+-------+                           //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +-------------+                            //
//           | BOARD_BusTakeFailure()        |/////////////|                            //
//           +-------------------------------+// FAILURE //|                            //
//                                           |/////////////|                            //
//                                           +-------------+                            //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat von IDLE nach TRANSMIT und beginnt  //
//  mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im     //
//  Gange (Zustand TRANSMIT), wird bei Aufruf der Funktion das zu übertragende Daten-   //
//  byte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach Aus-     //
//  sendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat    //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sich ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die    //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAILURE bleiben Aufrufe von BOARD_BusTransmit()   //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
        }
    } while (queueFull);
    // Byte senden
    Atomic
    {   if (busStatus.error) return; // Zustand FAILURE: Nichts tun
        if (busStatus.pending)
        {   // Zustand TRANSMIT: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            LL_USART_DisableIT_TC(USART1);
            LL_USART_TransmitData8(USART1, byte);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMIT befindet. Das ist solange der Fall,  //
//  wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {	pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in <byte> übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Bei einer Unterbrechung der Empfangssequenz kehrt die Funktion mit     //
//  'false' zurück. Umgekehrt gilt das jedoch nicht: Der Rückgabewert 'false' bedeutet  //
//  nicht zwingend, dass auch die Empfangssequenz bereits unterbrochen ist! Dazu siehe  //
//  BOARD_BusIsReceiving().                                                             //
//                                                                                      //
//  Eingabe: · byte: Referenz, die das nächste zur Abholung bereitstehende Datenbyte    //
//                   aufnimmt.                                                          //
//  Ausgabe: 'true', wenn in <byte> ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if (!busRxQueue[head].timeout)
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if (head != busRxQueueTail)
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
                head++;
                if (head == CARDINALITY(busRxQueue)) head = 0;
                busRxQueueHead = head;
                hasData = true;
            }
        }
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängende Empfangssequenz, wenn  //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als 'CFG_BUS_INTERBYTE_TIMEOUT'       //
//  Millisekunden auseinander (s.u. timeout), gilt die Empfangssequenz als unter-		//
//	brochen.     																		//
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//    Empfangssequenzen: ABC, DE, F...                                                  //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit 'true' zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                              //
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte)) // Datenbytes abholen und aufsammeln   //
//               {   if (index < sizeof(sequence))                                      //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//                index  = 0;                                                           //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if (head != busRxQueueTail) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
    busActivityIndicator = CFG_BUS_ACT_INDICATOR_DURATION;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusActivity(): Der Bus-Aktivitätsindikator zeigt an, ob aktuell mit dem       //
//  -------------------  Gerät kommuniziert wird. Diese Funktion liefert den Zustand    //
//  des Indikators. Nach Abschluss der Kommunikation bleibt der Indikator noch 'CFG_-   //
//  BUS_ACT_INDICATOR_DURATION' Millisekunden aktiviert.                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusActivity(void)
{
    return busActivityIndicator != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_PushButtonPressed(void)
{
	return !LL_GPIO_IsInputPinSet(GPIOA, LL_GPIO_PIN_0);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(Led led, bool on)
{
	switch (led)
	{   case LED_ONBOARD:
			if (on) LL_GPIO_SetOutputPin(GPIOB, LL_GPIO_PIN_3);
			else    LL_GPIO_ResetOutputPin(GPIOB, LL_GPIO_PIN_3);
			break;
	    case LED_BLUE:
			if (on) LL_GPIO_SetOutputPin(GPIOA, LL_GPIO_PIN_1);
			else    LL_GPIO_ResetOutputPin(GPIOA, LL_GPIO_PIN_1);
			break;
	    case LED_RED:
			if (on) LL_GPIO_SetOutputPin(GPIOA, LL_GPIO_PIN_3);
			else    LL_GPIO_ResetOutputPin(GPIOA, LL_GPIO_PIN_3);
			break;
		default:
			;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Led(Led led)
{
	switch (led)
	{   case LED_ONBOARD: return LL_GPIO_IsOutputPinSet(GPIOB, LL_GPIO_PIN_3);
		case LED_BLUE:    return LL_GPIO_IsOutputPinSet(GPIOA, LL_GPIO_PIN_1);
		case LED_RED:     return LL_GPIO_IsOutputPinSet(GPIOA, LL_GPIO_PIN_3);
		default:		  return false;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    DisableInterrupts();
    LL_RCC_ClearResetFlags();
    // Interrupt-Vektortabelle verlegen --------------------------------------------------
    {	extern char APPL_SECTION_START;		// Definiert im Linker Command File
    	SCB->VTOR = (UInt32)((volatile UInt32*)(&APPL_SECTION_START));
	}
    // Watchdog --------------------------------------------------------------------------
	{	IWDG->KR  = 0x0000CCCC;					// Watchdog aktivieren
		IWDG->KR  = 0x00005555; 				// Schreibzugriff auf PR, RLR und WINR
		IWDG->PR  = IWDG_PR_PR_1| IWDG_PR_PR_2;	// Vorteiler: 40 kHz ÷ 256 = 156 kHz
		IWDG->RLR = 2*156;						// Watchdog-Zeit auf 2 s
		while (IWDG->SR);						// Warten, bis Änderungen abschlossen sind
		IWDG->KR  = 0x0000AAAA;					// WD zurücksetzen/Schreibzugr. verbieten
	}
    // Systemtakte initialisieren --------------------------------------------------------
    {	LL_FLASH_SetLatency(LL_FLASH_LATENCY_0);
  		while(LL_FLASH_GetLatency()!= LL_FLASH_LATENCY_0);
  		LL_RCC_HSE_EnableBypass();
  		LL_RCC_HSE_Enable();
  		while(LL_RCC_HSE_IsReady() != 1);
  		LL_RCC_SetAHBPrescaler(LL_RCC_SYSCLK_DIV_1);
  		LL_RCC_SetAPB1Prescaler(LL_RCC_APB1_DIV_4);
  		LL_RCC_SetAPB2Prescaler(LL_RCC_APB2_DIV_4);
  		LL_RCC_SetSysClkSource(LL_RCC_SYS_CLKSOURCE_HSE);
  		while(LL_RCC_GetSysClkSource() != LL_RCC_SYS_CLKSOURCE_STATUS_HSE);
  		LL_Init1msTick(8000000);
  		LL_SetSystemCoreClock(8000000);
  		LL_RCC_SetUSARTClockSource(LL_RCC_USART1_CLKSOURCE_PCLK1);
	}
    // Busse aktivieren ------------------------------------------------------------------
    {	LL_APB2_GRP1_EnableClock(LL_APB2_GRP1_PERIPH_SYSCFG);
  		LL_APB1_GRP1_EnableClock(LL_APB1_GRP1_PERIPH_PWR);
  		LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOA);
  		LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOB);
  		LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOF);
		LL_APB2_GRP1_EnableClock(LL_APB2_GRP1_PERIPH_USART1);
	}
    // GPIOs -----------------------------------------------------------------------------
	{	// Taste: PA0
    	LL_GPIO_InitTypeDef setup = { 0 };
     	setup.Pin  = LL_GPIO_PIN_0;
     	setup.Mode = LL_GPIO_MODE_INPUT;
     	setup.Pull = LL_GPIO_PULL_UP;
     	LL_GPIO_Init(GPIOA, &setup);
	}
	{	// Blaue LED: PA1
    	LL_GPIO_InitTypeDef setup = { 0 };
     	setup.Pin        = LL_GPIO_PIN_1;
     	setup.Mode       = LL_GPIO_MODE_OUTPUT;
     	setup.Speed      = LL_GPIO_SPEED_FREQ_LOW;
     	setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
     	setup.Pull       = LL_GPIO_PULL_NO;
     	LL_GPIO_Init(GPIOA, &setup);
     	LL_GPIO_ResetOutputPin(GPIOA, LL_GPIO_PIN_1);
	}
	{	// Rote LED: PA3
    	LL_GPIO_InitTypeDef setup = { 0 };
     	setup.Pin        = LL_GPIO_PIN_3;
     	setup.Mode       = LL_GPIO_MODE_OUTPUT;
     	setup.Speed      = LL_GPIO_SPEED_FREQ_LOW;
     	setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
     	setup.Pull       = LL_GPIO_PULL_NO;
     	LL_GPIO_Init(GPIOA, &setup);
     	LL_GPIO_ResetOutputPin(GPIOA, LL_GPIO_PIN_3);
	}
	{	// Onboard-LED: PB3
		LL_GPIO_InitTypeDef setup = { 0 };
		setup.Pin        = LL_GPIO_PIN_3;
		setup.Mode       = LL_GPIO_MODE_OUTPUT;
		setup.Speed      = LL_GPIO_SPEED_FREQ_LOW;
		setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
		setup.Pull       = LL_GPIO_PULL_NO;
		LL_GPIO_Init(GPIOB, &setup);
		LL_GPIO_ResetOutputPin(GPIOB, LL_GPIO_PIN_3);
	}
	// Millisekunden-Zeitgeber -----------------------------------------------------------
	{	SystemCoreClockUpdate();
		SysTick_Config(SystemCoreClock/1000);
		NVIC_SetPriority(SysTick_IRQn, 0);
	}
	// USART 1 19200 Baud, 8 Data, 1 Stop, No Parity, Async. -----------------------------
	{	// USART1_TX: PA9
		{	LL_GPIO_InitTypeDef setup = { 0 };
	  		setup.Pin        = LL_GPIO_PIN_9;
	  		setup.Mode       = LL_GPIO_MODE_ALTERNATE;
	  		setup.Speed      = LL_GPIO_SPEED_FREQ_HIGH;
	  		setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
	  		setup.Pull       = LL_GPIO_PULL_NO;
	  		setup.Alternate  = LL_GPIO_AF_7;
	  		LL_GPIO_Init(GPIOA, &setup);
		}
	  	// USART1_RX: PA10
	  	{	LL_GPIO_InitTypeDef setup = { 0 };
	  		setup.Pin        = LL_GPIO_PIN_10;
			setup.Mode       = LL_GPIO_MODE_ALTERNATE;
			setup.Speed      = LL_GPIO_SPEED_FREQ_HIGH;
			setup.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
			setup.Pull       = LL_GPIO_PULL_NO;
			setup.Alternate  = LL_GPIO_AF_7;
			LL_GPIO_Init(GPIOA, &setup);
		}
	  	// USART1: Setup
		{	LL_USART_InitTypeDef setup = { 0 };
			setup.BaudRate            = 19200;
			setup.DataWidth           = LL_USART_DATAWIDTH_8B;
			setup.StopBits            = LL_USART_STOPBITS_1;
			setup.Parity              = LL_USART_PARITY_NONE;
			setup.TransferDirection   = LL_USART_DIRECTION_TX_RX;
			setup.HardwareFlowControl = LL_USART_HWCONTROL_NONE;
			setup.OverSampling        = LL_USART_OVERSAMPLING_16;
			LL_USART_Init(USART1, &setup);
			LL_USART_DisableOverrunDetect(USART1);
			LL_USART_DisableDMADeactOnRxErr(USART1);
			LL_USART_ConfigAsyncMode(USART1);
			LL_USART_EnableIT_RXNE(USART1);
			LL_USART_EnableIT_TC(USART1);
			LL_USART_Enable(USART1);
	  		NVIC_EnableIRQ(USART1_IRQn);
		}
	}
    // Sonstiges -------------------------------------------------------------------------
    // Hinweis zu Interrupt Prioritäten / Prioritätsgruppen / Subprioritäten:
    // Der Reset-Wert für die Aufteilung der Prioritäten in Prioritätsgruppen und Sub-
    // prioritäten (siehe NVIC_GetPriorityGrouping()) ist 0. Das bedeutet: Es gibt 16
    // Prioritätsgruppen aber keine Subprioritäten. Dazu PM0214, S.41:
    //    "Only the group priority determines preemption of interrupt exceptions. When the
    //    processor is executing an interrupt exception handler, another interrupt with the
    //    same group priority as the interrupt being handled does not preempt the handler."
    // Der Reset-Wert für die Interrupt Prioritäten (siehe NVIC_GetPriority(IRQn)) ist
    // ebenfalls 0. Damit haben alle IRQs die gleiche Prioritätsgruppe und unterbrechen
    // sich laut PM0214 gegenseitig nicht.
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
	ResetWatchdog();
	ProceedIndicators();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    while (BOARD_BusIsTransmitting());
    switch (doNext)
    {   case DO_ENTER_PGMLDR: EnterPgmLdr(); break;
        case DO_STARTUP:      Reset();       break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
