﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//      The Cy4NET System                      Version 19.10                            //
//      __________________________             _________________________________        //
//                                                                                      //
//      Decentralised Realtime                 Copyright (c) 2012 Reimar Grasbon        //
//      Peer-to-peer Communication             (www.cy4net.org)                         //
//      Protocol                                                                        //
//      _________________________              _________________________________        //
//                                                                                      //
//      [Communication Core Unit]                                                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Cy4NET-Datagramm                                                                    //
//  ----------------                                                                    //
//                                                                                      //
//                     0   1   2   3   4   5   6   7   8   9   10  ···      n+10        //
//                   +---+---+---+---+---+---+---+---+---+---+---+-···-+---+---+        //
//      Cy4NET-PDU:  |DPS|DTF|  DST  |  SRC  |    LOC    |CNT|     DAT     |CRC|        //
//                   +---+---+---+---+---+---+---+---+---+---+---+-···-+---+---+        //
//                                                                                      //
//                                                                                      //
//      · DPS:  Datagramm Einleitungs-Bitsequenz (Datagram Preamble Sequence): Dieses   //
//              immer gleiche 8-Bit Startmuster leitet die Übertragung einer Cy4-PDU    //
//              ein. Gesendet wird 10101010b (AAh).                                     //
//                                                                                      //
//      · DTF:  Datagramm Typflags (Datagram Type Flags): Die oberen 6 Bits enthalten   //
//              die Datagramm-Identifikationsbits DIB (Datagram Identification Bits).   //
//              Cy4-PDUs identifizieren sich hier stets mit der Bitfolge: 110000 (=     //
//              ASCII für '0'). Übrig bleiben Bit 1 und 0, die die Charakteristik des   //
//              Datagramms bestimmen:                                                   //
//                                                                                      //
//                            7 6 5 4 3 2 1 0                                           //
//                           +-+-+-+-+-+-+-+-+                                          //
//                           |1|1|0|0|0|0|D|P|                                          //
//                           +-+-+-+-+-+-+-+-+                                          //
//                            \_________/ | |                                           //
//                                DIB     | +---- P: Nutzlast-Flag: 1 = Führt Nutzlast  //
//                                        |          mit, 0 = keine Nutzlast.           //
//                                        +------ D: Datagramm-Richtungs-Flag (1 = An-  //
//                                                   frage, 0 = Antwort.                //
//                                                                                      //
//              Mit den Bits 'D' und 'P' lassen sich somit vier Wertausprägungen für    //
//              das DTF kombinieren. Sie legen den Typ des Datagramms fest:             //
//                                                                                      //
//                          DIB   DP | DTF-Wert | Typ                                   //
//                        -----------+----------+---------------                        //
//                         110000 10 |    C2h   | Leseanforderung                       //
//                         110000 01 |    C1h   | Leseantwort                           //
//                         110000 11 |    C3h   | Schreibaufforderung                   //
//                         110000 00 |    C0h   | Schreibbestätigung                    //
//                                                                                      //
//      · DST:  Cy4NET-Adresse des Empfängerknotens (siehe Cy4Address).                 //
//                                                                                      //
//      · SRC:  Cy4NET-Adresse des Sende-Knotens (siehe Cy4Address).                    //
//                                                                                      //
//      · LOC:  Dienstadresse. LOC ist eine vorzeichenlose 24-Bit-Zahl im Bereich von   //
//              0 bis FFFFFEh, FFFFFFh ist nicht erlaubt. Rundrufe sind festgelegt      //
//              auf den Bereich von E00000h bis EFFFFFh. Der Adressbereich unter        //
//              100000h ist für Zugriffe des Programmladermodules reserviert. LOC hat   //
//              die Byte-Reihenfolge Big Endian (High/Mid/Low)                          //
//                                                                                      //
//      · CNT:  Je nach Datagrammtyp, die Anzahl der                                    //
//                 · mitgeführten  (bei Schreibaufforderungen)                          //
//                 · geschriebenen (bei Schreibbestätigungen)                           //
//                 · angeforderten (bei Leseanforderungen)                              //
//                 · gelesenen     (bei Leseantworten)                                  //
//              Datenbytes. Wertebereich: 1 .. 64                                       //
//                                                                                      //
//      · DAT:  Nutzlast. Datagramme vom Typ Schreibaufforderung und Leseantwort        //
//              führen in dieser Sektion <CNT> Datenbytes mit sich (siehe 'P' Flag im   //
//              DTF). Bei Schreibbestätigungen und Leseanforderungen ist die Sektion    //
//              DAT leer.                                                               //
//                                                                                      //
//      · CRC:  8-Bit Dow-CRC über alle Datagrammbytes, außer CRC.                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "ARMChip.h"
#include "Cy4Core.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define  CY4_PAYLOAD_CAPACITY     64  // Nutzlast-Kapazität (feste Größe) [Bytes]
#define  CY4_ROUND_TRIP_TIMEOUT  500  // Paketumlauf-Zeitgrenze:          300 .. 500 [ms]
#define  CY4_QUEUING_DELAY        50  // Neusende-Verzögerung nach Übertr.: 5 .. 100 [ms]
#define  CY4_BACKOFF_MIN           3  // Min. Verzögerung vor Neuversuch:   1 ..   5 [ms]
#define  CY4_BACKOFF_MAX          15  // Max. Verzögerung vor Neuversuch:   5 ..  20 [ms]
#define  PDU_DPS                0xAA  // Datagramm Einleitungs-Bitsequenz: 0b10101010
#define  PDU_DIB                0x30  // Datagramm Identifikationsbits: 0b110000
#define  PDU_DTF_P              BV(0) // Datagramm-Typflag: P (Payload)
#define  PDU_DTF_D              BV(1) // Datagramm-Typflag: D (Direction)
#define  USE_CRC_LOOKUP_TABLE   true  // 'true' = CRC-Tabelle, 'false' CRC-Berechnung

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct                  // Datagramm
{   UInt08  byte[10+CY4_PAYLOAD_CAPACITY+1];
} PDU;

//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                    // Übertragungs-Zustandsautomat:
{   ST_RECEIVE      = 0,        // · Warten auf Empfang eines Datagramms
    ST_INBOX        = 1,        // · Ein Datagramm liegt zur Abholung im Eingangsfach
    ST_IDLE         = 2,        // · Warten auf Übertragungs-Auftrag
    ST_OUTBOX       = 3,        // · Ein Datagramm liegt zur Übertragung im Ausgangsfach
    ST_TRANSMIT     = 4         // · Datagramm wird übertragen
} State;

//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                    // Ereignisse (Auslöser) des Zustandsautomaten:
{   EV_NONE         = 0,        // · Kein Ereignis
    EV_RECEIVED     = BV(0),    // · Neues Datagramm ist eingetroffen
    EV_HANDLED      = BV(1),    // · Datagramm wurde verarbeitet oder neu eingestellt
    EV_ADMISSION    = BV(2),    // · Freigabe zur Übertragung des Datagramms erteilt
    EV_TRANSMITTED  = BV(3),    // · Übertragung des Datagramms abgeschlossen
    EV_COLLISION    = BV(4),    // · Übertragungsfehler bzw. Kollision aufgetreten
    EV_TIMEOUT      = BV(5),    // · Paketumlaufzeit abgelaufen
    EV_INVALID      = BV(6),    // · Datagramm ist ungültig
    EV_BROADCAST    = BV(7)     // · Datagramm ist ein Rundruf
} Event;

//////////////////////////////////////////////////////////////////////////////////////////
typedef struct                  // Zeitgeber:
{   UInt16  ticks;              // · Anzahl Millisekunden-Ticks. Ein Wert von 0 bedeutet:
} TTimer;                       //   Zeitgeber ist abgelaufen.

//////////////////////////////////////////////////////////////////////////////////////////
typedef struct                  // Ereignis-Satz:
{   UInt08  set;                // · Oder-verknüpfte 'Event'-Einträge
} EventSet;

//////////////////////////////////////////////////////////////////////////////////////////
typedef void (*Action)(void);   // Aktionsfunktion

//////////////////////////////////////////////////////////////////////////////////////////
typedef struct                  // Wertetabellen-Eintrag der Zustandsübergangsfunktion
{   Event   event;              // · Auslösendes Ereignis
    Action  action;             // · Auszuführende Aktion
    State   state;              // · Folgezustand
} ZetaItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Cy4Address selfAddress    = CY4ADDRESS_INVALID; // Eigene Cy4-Adresse
static PDU        receivePdu     = { { 0 } };          // Empfangenes Datagramm
static bool       newPduReceived = false;              // Neues Datagramm-Flag
///////////////////////////////////////////////////////// Eingangs-Übertragungseinheit:
static State      inState        = ST_RECEIVE;         // · Zustand des Eingangs-Automaten
static EventSet   inEventSet     =   { 0 };            // · Satz von Ereignissen
static TTimer     inTimer        =   { 0 };            // · Paketumlauf-Zeitgeber
static TTimer     inPause        =   { 0 };            // · Pausen-Zeitgeber
static PDU        inPdu          = { { 0 } };          // · Datagramm
///////////////////////////////////////////////////////// Ausgangs-Übertragungseinheit:
static State      outState       = ST_IDLE;            // · Zustand des Ausgangs-Automaten
static EventSet   outEventSet    =   { 0 };            // · Satz von Ereignissen
static TTimer     outTimer       =   { 0 };            // · Paketumlauf-Zeitgeber
static TTimer     outPause       =   { 0 };            // · Pausen-Zeitgeber
static PDU        outPdu         = { { 0 } };          // · Datagramm

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Invalidate(): Erklärt das angegebene Datagramm für ungültig. Wichtig: Der Wechsel   //
//  ------------  zur Ungültigkeit ist lediglich ein Statuswechsel. Bis auf das DPS     //
//  bleibt der Inhalt eines Datagramms bei einer Ungültigkeitserklärung unangetastet.   //
//                                                                                      //
//  Eingabe: · pdu: Das für ungültig zu erklärende Datagramm.                           //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Invalidate(PDU* pdu)
{
    pdu->byte[0] = !PDU_DPS;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsValid(): Prüft, ob das angegebene Datagramm den Status 'gültig' hat. Siehe auch   //
//  ---------  Invalidate().                                                            //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: 'true' wenn das Datagramm den Status gültig hat.                           //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsValid(const PDU* pdu)
{
    return pdu->byte[0] == PDU_DPS;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsRequest(): Prüft, ob es sich bei dem angegebenen Datagramm um eine Anfrage        //
//  -----------  handelt, andernfalls ist es eine Antwort.                              //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: 'true', wenn das Datagramm eine Anfrage ist.                               //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsRequest(const PDU* pdu)
{
    return pdu->byte[1] & PDU_DTF_D;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsBroadcast(): Prüft, ob es sich bei dem angegebenen Datagramm um einen Rundruf     //
//  -------------  handelt. Ein Datagramm wird zu einem Rundruf, wenn mindestens eine   //
//  Komponente der Zieladresse, d.h. Netz- oder Geräteadresse, 0 ist.                   //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: 'true', wenn das Datagramm ein Rundruf ist.                                //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsBroadcast(const PDU* pdu)
{
    return pdu->byte[2] == 0 || pdu->byte[3] == 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsExxxxxWriteRequest(): Prüft, ob das angegebene Datagramm eine Schreibaufforderung //
//  ----------------------  im Dienstadressbereich von E00000h bis EFFFFFh ist. Neben   //
//  der in IsBroadcast() angegebenen Eigenschaft ist das die zusätzliche Bedingung für  //
//  einen gültigen Rundruf.                                                             //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: 'true', wenn das Datagramm eine Schreibaufforderung im Bereich Exxxxx ist. //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsExxxxxWriteRequest(const PDU* pdu)
{
    return pdu->byte[1] == 0xC3 && (pdu->byte[6] & 0xF0) == 0xE0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Access(): Ermittelt die Zugriffsart (d.h. lesend oder schreibend) des angegebenen   //
//  --------  Datagramms. Die Funktion liefert eine Ausprägung des Aufzählungstyps      //
//  'Cy4MessageAccess' (siehe dort).                                                    //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: CY4MESSAGE_READ oder CY4MESSAGE_WRITE.                                     //
//////////////////////////////////////////////////////////////////////////////////////////
static Cy4MessageAccess
Access(const PDU* pdu)
{
    if (pdu->byte[1] == 0xC1 || pdu->byte[1] == 0xC2) return CY4MESSAGE_READ;
    else                                              return CY4MESSAGE_WRITE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  HasPayload(): Prüft, ob das angegebene Datagramm Nutzlast mitführt. Nur Datagramme  //
//  ------------  vom Typ Schreibaufforderung und Leseantwort führen Nutzlast mit sich. //
//                                                                                      //
//  Eingabe: · pdu: Das zu prüfende Datagramm.                                          //
//  Ausgabe: 'true', Datagramm führt Nutzlast mit sich.                                 //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
HasPayload(const PDU* pdu)
{
    return (bool)(pdu->byte[1] & PDU_DTF_P);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  AddEvent(): Fügt dem angegebenen Ereignis-Satz ein Ereignis hinzu.                  //
//  ----------                                                                          //
//                                                                                      //
//  Eingabe: · eventSet: Der zu erweiternde Ereignis-Satz.                              //
//           · event:    Ereignis, das hinzugefügt werden soll.                         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
AddEvent(EventSet* eventSet, Event event)
{
    eventSet->set |= event;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  HasEvent(): Prüft, ob das angegebene Ereignis im angegebenen Ereignis-Satz ent-     //
//  ----------  halten ist.                                                             //
//                                                                                      //
//  Eingabe: · eventSet: Der zu prüfende Ereignis-Satz.                                 //
//           · event:    Das vorkommende Ereignis.                                      //
//  Ausgabe: 'true', wenn das Ereignis in <eventSet> vorkommt.                          //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
HasEvent(const EventSet* eventSet, Event event)
{
    return (bool)(eventSet->set & event);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ClearEvents(): Entfernt alle Ereignisse aus dem angegebenen Ereignis-Satz.          //
//  -------------                                                                       //
//                                                                                      //
//  Eingabe: · eventSet: Der zu löschende Ereignis-Satz.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ClearEvents(EventSet* eventSet)
{
    eventSet->set = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  StartTimer(): Stellt den Zeitgeber in <timer> auf <duration> Millisekunden. Läuft   //
//  ------------  der Zeitgeber bereits, wird er zurückgesetzt und neu gestartet.       //
//                                                                                      //
//  Eingabe: · timer:    Adresse des zu startenden Zeitgebers.                          //
//           · duration: Dauer bis zur Zeitüberschreitung: 1 .. 255 [ms].               //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
StartTimer(TTimer* timer, UInt08 duration)
{
    timer->ticks = BOARD_MSecTicks() + (UInt16)duration;
    if (timer->ticks == 0) timer->ticks = 1;   // Wert 0 ist reserviert
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsTimeout(): Prüft, ob der Zeitgeber in <timer> bereits abgelaufen ist. Die Zeitab- //
//  -----------  stände, in denen bei laufendem Zeitgeber IsTimeout() abgefragt wird,   //
//  dürfen nicht länger als eine Minute auseinanderliegen.                              //
//                                                                                      //
//  Eingabe: · timer: Adresse des abzufragenden  Zeitgebers.                            //
//  Ausgabe: 'true', wenn der Zeitgeber abgelaufen ist.                                 //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsTimeout(TTimer* timer)
{
    if (timer->ticks != 0)
    {   UInt16 ticks = BOARD_MSecTicks();
        if ((SInt16)(timer->ticks - ticks) > 0) return false;
        timer->ticks = 0;   // 0 ist magischer Wert für "abgelaufen"
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  HasNewPdu(): Prüft, ob in der Empfangseinheit ein neues Datagramm eingetroffen ist. //
//  -----------                                                                         //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein neues Datagramm vorliegt.                                 //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
HasNewPdu(void)
{
    return newPduReceived;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TakeNewPdu(): Übernimmt das neu empfangene Datagramm in <pdu> in gibt den Empfang   //
//  ------------  weiterer Datagramme wieder frei. TakeNewPdu() prüft nicht, ob über-   //
//  haupt ein neues Datagramm vorlag (dazu siehe HasNewPdu()).                          //
//                                                                                      //
//  Eingabe: pdu: Referenz, die das neue Datagramm aufnimmt.                            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
TakeNewPdu(PDU* pdu)
{
    *pdu = receivePdu;
    newPduReceived = false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DropNewPdu(): Verwirft das neu empfangene Datagramm und gibt den Empfang weiterer   //
//  ------------  Datagramme wieder frei. DropNewPdu() prüft nicht, ob überhaupt ein    //
//  neues Datagramm vorlag (dazu siehe HasNewPdu()).                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
DropNewPdu(void)
{
    newPduReceived = false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  UpdateCRC(): Ermittelt den Wert der zyklischen Redundanzprüfung (CRC, 8-Bit)        //
//  -----------  für den Bytewert <data>. Dabei wird der (neue) CRC aufbauend auf den   //
//  übergebenen Wert in <crc> berechnet.                                                //
//  Das zu Grunde liegende Polynom ist X^8+X^5+X^4+1, auch bekannt als "DOW CRC". Die   //
//  Berechnungsmethode liefert das folgende Schema:                                     //
//                                                                         inBit        //
//                                                                           | ____     //
//    +---+  +---+  +---+  +---+   ____   +---+   ____   +---+  +---+  +---+ +-\XOR\    //
// +->|MSB|->|   |->|   |->|   |---\XOR\--|   |---\XOR\--|   |->|   |->|LSB|---/___/-+  //
// |  +---+  +---+  +---+  +---+ +-/___/  +---+ +-/___/  +---+  +---+  +---+         |  //
// |                             |              |                                    |  //
// +-----------------------------+--------------+------------------------------------+  //
//                                                                                      //
//                                                                                      //
//  Eingabe: · data: Wert, der zu <crc> hinzugerechnet werden soll.                     //
//           · crc:  Referenz, die den bisherigen CRC-Wert bereitstellt und den neuen   //
//                   wieder aufnimmt.                                                   //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
UpdateCRC(UInt08 data, UInt08* crc)
{

    #if USE_CRC_LOOKUP_TABLE == true  // /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

    // Ermittlung des CRCs mittels Nachschlagetabelle
    static const UInt08 table[] PROGMEM =
    {   0x00, 0x5E, 0xBC, 0xE2, 0x61, 0x3F, 0xDD, 0x83, 0xC2, 0x9C, 0x7E, 0x20, 0xA3,
        0xFD, 0x1F, 0x41, 0x9D, 0xC3, 0x21, 0x7F, 0xFC, 0xA2, 0x40, 0x1E, 0x5F, 0x01,
        0xE3, 0xBD, 0x3E, 0x60, 0x82, 0xDC, 0x23, 0x7D, 0x9F, 0xC1, 0x42, 0x1C, 0xFE,
        0xA0, 0xE1, 0xBF, 0x5D, 0x03, 0x80, 0xDE, 0x3C, 0x62, 0xBE, 0xE0, 0x02, 0x5C,
        0xDF, 0x81, 0x63, 0x3D, 0x7C, 0x22, 0xC0, 0x9E, 0x1D, 0x43, 0xA1, 0xFF, 0x46,
        0x18, 0xFA, 0xA4, 0x27, 0x79, 0x9B, 0xC5, 0x84, 0xDA, 0x38, 0x66, 0xE5, 0xBB,
        0x59, 0x07, 0xDB, 0x85, 0x67, 0x39, 0xBA, 0xE4, 0x06, 0x58, 0x19, 0x47, 0xA5,
        0xFB, 0x78, 0x26, 0xC4, 0x9A, 0x65, 0x3B, 0xD9, 0x87, 0x04, 0x5A, 0xB8, 0xE6,
        0xA7, 0xF9, 0x1B, 0x45, 0xC6, 0x98, 0x7A, 0x24, 0xF8, 0xA6, 0x44, 0x1A, 0x99,
        0xC7, 0x25, 0x7B, 0x3A, 0x64, 0x86, 0xD8, 0x5B, 0x05, 0xE7, 0xB9, 0x8C, 0xD2,
        0x30, 0x6E, 0xED, 0xB3, 0x51, 0x0F, 0x4E, 0x10, 0xF2, 0xAC, 0x2F, 0x71, 0x93,
        0xCD, 0x11, 0x4F, 0xAD, 0xF3, 0x70, 0x2E, 0xCC, 0x92, 0xD3, 0x8D, 0x6F, 0x31,
        0xB2, 0xEC, 0x0E, 0x50, 0xAF, 0xF1, 0x13, 0x4D, 0xCE, 0x90, 0x72, 0x2C, 0x6D,
        0x33, 0xD1, 0x8F, 0x0C, 0x52, 0xB0, 0xEE, 0x32, 0x6C, 0x8E, 0xD0, 0x53, 0x0D,
        0xEF, 0xB1, 0xF0, 0xAE, 0x4C, 0x12, 0x91, 0xCF, 0x2D, 0x73, 0xCA, 0x94, 0x76,
        0x28, 0xAB, 0xF5, 0x17, 0x49, 0x08, 0x56, 0xB4, 0xEA, 0x69, 0x37, 0xD5, 0x8B,
        0x57, 0x09, 0xEB, 0xB5, 0x36, 0x68, 0x8A, 0xD4, 0x95, 0xCB, 0x29, 0x77, 0xF4,
        0xAA, 0x48, 0x16, 0xE9, 0xB7, 0x55, 0x0B, 0x88, 0xD6, 0x34, 0x6A, 0x2B, 0x75,
        0x97, 0xC9, 0x4A, 0x14, 0xF6, 0xA8, 0x74, 0x2A, 0xC8, 0x96, 0x15, 0x4B, 0xA9,
        0xF7, 0xB6, 0xE8, 0x0A, 0x54, 0xD7, 0x89, 0x6B, 0x35
    };

    *crc = table[data^(*crc)];

    #else // /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

    // Ermittlung des CRCs durch Rechnung (4 mal langsamer aber halber Speicherverbrauch)

    *crc ^= data;
    for (UInt08 i=0; i<8; i++) *crc = (*crc >> 1) ^ ((*crc & 1) ? 0x8C : 0 );

    #endif  // /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Transmit(): Sendet das Datagramm über den Bus.                                      //
//  ----------                                                                          //
//                                                                                      //
//  Eingabe: · pdu: Das zu übertragende Datagramm.                                      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Transmit(const PDU* pdu)
{
    UInt08  n;

    if (HasPayload(pdu)) n = 11 + pdu->byte[9];
    else                 n = 11;
    for (UInt08 i=0; i<n; i++) BOARD_BusTransmit(pdu->byte[i]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  StartBackoff(): Gab es beim Senden des Datagramms eine Bus-Kollision oder einen     //
//  --------------  Fehler, stoppt der Sendevorgang unmittelbar. Nach einer kurzen      //
//  Pause versucht die Übertragungseinheit das Datagramm erneut auf die Reise zu        //
//  schicken. Diese Neuversuchs-Verzögerung wird "Backoff" genannt. Der Backoff hier    //
//  ist eine zufällig gewählte Zeitspanne zwischen CY4_BACKOFF_MIN und CY4_BACKOFF_MAX  //
//  Millisekunden. StartBackoff() startet den angegebenen Timer mit dieser Verzögerung. //
//  Die Funktion wird eingesetzt in Action_StartRequestBackoff() und Action_StartRes-   //
//  ponseBackoff().                                                                     //
//                                                                                      //
//  Eingabe: · timer: Zeitgeber, der mit der Backoff-Verzögerung gestartet wird.        //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
StartBackoff(TTimer* timer)
{
    static UInt08  backoff = CY4_BACKOFF_MIN;

    backoff += 1 + (BOARD_MSecTicks()&0x03);
    if (backoff > CY4_BACKOFF_MAX) backoff = CY4_BACKOFF_MIN;
    StartTimer(timer, backoff);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  StartQueuingDelay(): Nach Abschluss einer erfolgreichen Übertragung (Anfrage&Ant-   //
//  -------------------  wort), muss das Übertragungssystem (genauer das Ausgangs-      //
//  Übertragungssystem) eine kurze Pause einlegen, bevor es eine neue Anfrage auf die   //
//  Reise schicken darf. Diese Neusende-Verzögerung ("Queuing Delay") verhindert, dass  //
//  ein einzelner Knoten durch sukzessive Sendewünsche den Bus koplett belegt und       //
//  andere Knoten in ihrer Kommunikation blockiert werden.                              //
//  StartQueuingDelay() wird von den Aktionsfunktionen Action_VirtualResponse() und     //
//  Action_AcquireResponse() aufgerufen. Sie stellt den Ausgangs-Pausen-Zeitgeber auf   //
//  CY4_QUEUING_DELAY Millisekunden.                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
StartQueuingDelay(void)
{
    StartTimer(&outPause, CY4_QUEUING_DELAY);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Responsify(): Verwandelt das Anfrage-Datagramm des Ausgangs-Übertragungssystem in   //
//  ------------  ein Antwort-Datagramm. Das ist nötig, wenn das System keine Antwort   //
//  über den Bus erhalten hat und sich selbst eine Antwort kreieren muss. Responsify()  //
//  wird in den Aktionsfunktionen Action_InvalidateResponse() und Action_VirtualRes-    //
//  ponse() eingesetzt.                                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Responsify(void)
{
    // Datagramm von Anfrage in Antwort umwandeln
    outPdu.byte[1] = (PDU_DIB<<2) | (0x03 & ~outPdu.byte[1]);
    // Quell- und Zieladresse tauschen
    outPdu.byte[4] = outPdu.byte[2];
    outPdu.byte[5] = outPdu.byte[3];
    outPdu.byte[2] = selfAddress.net;
    outPdu.byte[3] = selfAddress.dev;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_InvalidateResponse(): Aktionsfunktion des Zustandsautomaten für ausgehende   //
//  ---------------------------  Nachrichten. Die Funktion wird aufgerufen, wenn das    //
//  Ausgangs-Übertragungssystem auf eine Anfrage keine Antwort erhalten hat bzw. wenn   //
//  die Zeitgrenze für den Erhalt der Antwort überschritten wurde. In dem Fall erzeugt  //
//  das System selbst eine Antwort, indem es die Anfrage in eine Antwort umwandelt. Die //
//  erzeugte Antwort ist jedoch als ungültig gekennzeichnet.                            //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_InvalidateResponse(void)
{
    Responsify();
    Invalidate(&outPdu);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_VirtualResponse(): Aktionsfunktion des Zustandsautomaten für ausgehende      //
//  ------------------------  Nachrichten. Die Funktion wird aufgerufen, wenn das Aus-  //
//  gangs-Übertragungssystem eine Rundruf-Anfrage abgesendet hat. Rundrufe sind immer   //
//  Schreibaufforderungen und bleiben immer unbeantwortet. Daher erzeugt das System die //
//  nötige Schreibbestätigung künstlich, indem es die Rundrufanfrage in eine Antwort    //
//  umwandelt. Das geht problemlos, da die Antwort auf eine Schreibaufforderung keine   //
//  Nutzlast befördert, sie ist lediglich die Quittung für die Anfrage. Ob der Rundruf  //
//  letztlich alle betroffenen Knoten erreicht, bleibt unbekannt. Da die Kommunika-     //
//  tionssequenz damit nominell abgeschlossen ist, wird nach Erzeugung der künstlichen  //
//  Antwort die Neusende-Pause eingelegt (dazu siehe StartQueuingDelay()).              //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_VirtualResponse(void)
{
    Responsify();
    StartQueuingDelay();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_AcquireRequest(): Aktionsfunktion des Zustandsautomaten für eingehende       //
//  -----------------------  Nachrichten. Die Funktion wird aufgerufen, wenn das Ein-   //
//  gangs-Übertragungssystem eine Anfrage erhalten hat. Die Anfrage wird entgegenge-    //
//  nommen und der Zeitgeber für die Verarbeitung der Anfrage und das Senden der Ant-   //
//  wort gestellt. Die Zeitgrenze ist die halbe Paketumlaufzeit, da bereits für das     //
//  Versenden der Nachricht auf Seiten des Quell-Knotens die halbe Paketumlaufzeit ver- //
//  anschlagt wurde.                                                                    //
//  Zusätzlich wird Busaktivität signalisiert.                                          //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_AcquireRequest(void)
{
    TakeNewPdu(&inPdu);
    StartTimer(&inTimer, CY4_ROUND_TRIP_TIMEOUT/2);
    BOARD_BusOnActivity();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_AcquireResponse(): Aktionsfunktion des Zustandsautomaten für ausgehende      //
//  ------------------------  Nachrichten. Die Funktion wird aufgerufen, wenn das Aus-  //
//  gangs-Übertragungssystem eine Antwort auf eine zuvor verschickte Anfrage erhalten   //
//  hat. Da das der korrekte Abschluss einer Kommunikationssequenz ist, wird nach ent-  //
//  gegennahme der Antwort die Neusende-Pause eingelegt (dazu siehe StartQueuing-       //
//  Delay()).                                                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_AcquireResponse(void)
{
    TakeNewPdu(&outPdu);
    StartQueuingDelay();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_StartTimer(): Aktionsfunktion des Zustandsautomaten für ausgehende Nach-     //
//  -------------------  richten. Durch Aufruf der Funktion wird eine Zeitgrenze für    //
//  das Senden einer Anfrage ebenso wie das Empfangen einer Antwort gesetzt. Daher      //
//  beträgt die Zeitgrenze auch jeweils die Hälfte der Paketumlaufzeit.                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_StartTimer(void)
{
    StartTimer(&outTimer, CY4_ROUND_TRIP_TIMEOUT/2);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_StartRequestBackoff(): Aktionsfunktion des Zustandsautomaten für ausgehende  //
//  ----------------------------  Nachrichten. Kam es während des Sendens einer Anfrage //
//  zu einer Bus-Kollision oder zu einem Fehler, wird die Anfrage nach Einhaltung einer //
//  Pause wiederholt. Dazu siehe StartBackoff().                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_StartRequestBackoff(void)
{
    StartBackoff(&outPause);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_StartResponseBackoff(): Aktionsfunktion des Zustandsautomaten für eingehende //
//  -----------------------------  Nachrichten. Kam es während des Sendens einer Ant-   //
//  wort zu einer Bus-Kollision oder zu einem Fehler, wird der Sendevorgang nach Ein-   //
//  haltung einer Pause wiederholt. Dazu siehe StartBackoff().                          //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_StartResponseBackoff(void)
{
    StartBackoff(&inPause);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_TransmitRequest(): Aktionsfunktion des Zustandsautomaten für ausgehende      //
//  ------------------------  Nachrichten. Die Aktionsfunktion sendet ein Anfrage-Data- //
//  gramm über den Bus.                                                                 //
//  Zusätzlich wird Busaktivität signalisiert                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_TransmitRequest(void)
{
    Transmit(&outPdu);
    BOARD_BusOnActivity();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Action_TransmitResponse(): Aktionsfunktion des Zustandsautomaten für eingehende     //
//  -------------------------  Nachrichten. Die Aktionsfunktion sendet ein Antwort-     //
//  Datagramm über den Bus.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
Action_TransmitResponse(void)
{
    Transmit(&inPdu);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedReceiving(): Verfahrensfunktion der Datagramm-Empfangseinheit. Die Funktion  //
//  ------------------  nimmt empfangene Datenbytes entgegen und versucht daraus ein    //
//  gültiges Datagramm zusammenzusetzen. Im Erfolgsfall (Abfrage durch HasNewPdu())     //
//  wird das Datagramm der Umgebung zur Verfügung gestellt und der Empfang weiterer     //
//  Datagramme so lange blockiert, bis die Umgebung das Datagramm als verarbeitet       //
//  (TakeNewPdu()) oder verworfen (DropNewPdu()) gekennzeichnet hat.                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedReceiving(void)
{
    if (newPduReceived) return; // Ein zuvor empf. Datagramm noch nicht verarbeitet
    static UInt08  index = 0;
    if (!BOARD_BusIsReceiving()) index = 0; // Datenstrom unterbrochen, zurück auf Anfang
    UInt08 byte;
    while (BOARD_BusReceive(&byte))
    {   static UInt08  crc = 0;
        static UInt08  len = 0;
        switch (index)
        {   case 0: // Prüfen, ob das erste Byte die "Datagramm Preamble Sequence" ist
                if (byte != PDU_DPS) continue;
                len = 10;
                crc = 0xFF;
                break;
            case 1: // Prüfen, ob das zweite Byte gültige "Datagramm Type Flags" hat
                if ((byte>>2) != PDU_DIB)
                {   index = 0;
                    continue;
                }
                break;
            case 9: // Nutzlast-Grenze prüfen und Datagrammlänge bestimmen
                if (byte == 0 || byte > CY4_PAYLOAD_CAPACITY)
                {   index = 0;
                    continue;
                }
                if (HasPayload(&receivePdu)) len += byte;
                break;
            default:
                ;
        }
        receivePdu.byte[index] = byte;
        UpdateCRC(byte, &crc);
        if (index == len)
        {   // Datagrammempfang abgeschlossen
            index = 0;
            if (crc == 0)
            {   // Prüfsumme ist korrekt
                if ((receivePdu.byte[2] == selfAddress.net || receivePdu.byte[2] == 0) &&
                    (receivePdu.byte[3] == selfAddress.dev || receivePdu.byte[3] == 0) &&
                    (receivePdu.byte[4] != 0               && receivePdu.byte[5] != 0) &&
                    (!IsBroadcast(&receivePdu) || IsExxxxxWriteRequest(&receivePdu)))
                {   // Datagramm ist direkt an uns adressiert oder ein gültiger Rundruf
                    newPduReceived = true;
                }
            }
            break;
        }
        index++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  EventGenerator(): Erzeugt die für die Umsetzung der Zustandsautomaten nötigen Er-   //
//  ----------------  eignisse:                                                         //
//                                                                                      //
//            · EV_RECEIVED:    Neues Datagramm ist eingetroffen.                       //
//            · EV_HANDLED:     s.u.                                                    //
//            · EV_ADMISSION:   Freigabe zur Übertragung des Datagramms erteilt.        //
//            · EV_TRANSMITTED: Übertragung des Datagramms abgeschlossen.               //
//            · EV_COLLISION:   Übertragungsfehler bzw. Kollision aufgetreten.          //
//            · EV_TIMEOUT:     Paketumlaufzeit abgelaufen.                             //
//            · EV_INVALID:     Datagramm ist ungültig.                                 //
//            · EV_BROADCAST:   Datagramm ist ein Rundruf.                              //
//                                                                                      //
//  Ausnahme ist das Ereignis EV_HANDLED, das von den Ablieferungs- und Abholungsfunk-  //
//  tionen (CY4CORE_Post...() und CY4CORE_Take...()) direkt ausgelöst wird.             //
//                                                                                      //
//  EventGenerator() wird aus den Verfahrensfunktionen der Zustandsautomaten aufge-     //
//  rufen.                                                                              //
//                                                                                      //
//  Eingabe: · eventSet: Referenz auf ein Ereigbnis-Satz, der die erzeugten Ereignisse  //
//                       aufnimmt.                                                      //
//           · timer:    Paketumlauf-Zeitgeber.                                         //
//           · pause:    Pausen-Zeitgeber (Neusende- oder Neuversuch-Verzögerung).      //
//           · pdu:      Betroffenes Datagramm.                                         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
EventGenerator(EventSet* eventSet, TTimer* timer, TTimer* pause, const PDU* pdu)
{
    // Ereignis EV_RECEIVED
    if (HasNewPdu() )
    {   // Neues Datagramm erhalten
        if (IsRequest(&receivePdu))
        {   // Neues Datagramm ist Anfrage
            AddEvent(&inEventSet, EV_RECEIVED);
        }
        else
        {   // Neues Datagramm ist Antwort
            if (outState != ST_TRANSMIT)
            {   bool utilize = false;
                if (outState == ST_RECEIVE)
                {   // Prüfen, ob das Datagramm die passende Antwort ist
                    if ((receivePdu.byte[1]&PDU_DTF_P) == (~outPdu.byte[1]&PDU_DTF_P))
                    {   static const UInt08 cmpIndex[][2] PROGMEM = { {2,4},{3,5},{6,6},
                                                                      {7,7},{8,8},{9,9}
                                                                    };
                        UInt08 i;
                        for (i=0; i<CARDINALITY(cmpIndex); i++)
                        {   UInt08 a = cmpIndex[i][0];
                            UInt08 b = cmpIndex[i][1];
                            if (receivePdu.byte[a] != outPdu.byte[b]) break;
                        }
                        if (i == CARDINALITY(cmpIndex)) utilize = true;
                    }
                }
                if (utilize) AddEvent(&outEventSet, EV_RECEIVED);
                else         DropNewPdu();   // Empfangene Antwort verwerfen
            }
        }
    }
    // Ereignis EV_ADMISSION               // Wenn:
    if (inState  != ST_TRANSMIT &&         // 1. Keine eingehende Übertragung im Gange
        outState != ST_TRANSMIT &&         // 2. Keine ausgehende Übertragung im Gange
        BOARD_BusIsIdle()       &&         // 3. Bus ist frei
        IsTimeout(pause))                  // 4. Keine Neusende- oder Neuversuchs-Pause
    {   AddEvent(eventSet, EV_ADMISSION);  // dann: Freigabe zur Übertragung erteilen
    }
    // Ereignis EV_TIMEOUT
    if (IsTimeout(timer)) AddEvent(eventSet, EV_TIMEOUT);
    // Ereignis EV_TRANSMITTED oder EV_COLLISION
    if (eventSet == &inEventSet)
    {   // Eingehende Nachrichten
        if (inState == ST_TRANSMIT)
        {   if (!BOARD_BusIsTransmitting())
            {   if (BOARD_BusTakeFailure()) AddEvent(eventSet, EV_COLLISION);
                else                        AddEvent(eventSet, EV_TRANSMITTED);
            }
        }
    }
    else
    {   // Ausgehende Nachrichten
        if (outState == ST_TRANSMIT)
        {   if (!BOARD_BusIsTransmitting())
            {   if (BOARD_BusTakeFailure()) AddEvent(eventSet, EV_COLLISION);
                else                        AddEvent(eventSet, EV_TRANSMITTED);
            }
        }
    }
    // Ereignis EV_BROADCAST
    if (IsBroadcast(pdu)) AddEvent(eventSet, EV_BROADCAST);
    // Ereignis EV_INVALID
    if (!IsValid(pdu)) AddEvent(eventSet, EV_INVALID);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  StateTransition(): Ermittelt aus dem aktuellen Zustand und den Ereignisse den       //
//  -----------------  Folgezustand und führt die dazu passende Aktion aus. Grundlage   //
//  ist die Zustandsübergangsfunktion Zeta, die hier in Form einer Wertetabelle gegeben //
//  ist. Zeta ist definiert als:                                                        //
//                                                                                      //
//                         Zeta:  State × Event  ->  State × Action                     //
//                                                                                      //
//  StateTransition() wird aus den Verfahrensfunktionen der Zustandsautomaten aufge-    //
//  rufen.                                                                              //
//                                                                                      //
//  Eingabe: · state:    Referenz, die den aktuellen Zustand bereitstellt und den       //
//                       Folgezustand wieder aufnimmt.                                  //
//           · eventSet: Satz mit Ereignissen.                                          //
//           · zeta:     Wertetabelle der Übergangsfunktion.                            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
StateTransition(State* state, const EventSet* eventSet, const ZetaItem (*zeta)[][4])
{
    Action action = NULL;
    UInt08 i=0;
    while (i<CARDINALITY((*zeta)[0]))
    {   Event event = (*zeta)[(UInt08)*state][i].event;
        if (event == EV_NONE) break;
        if (HasEvent(eventSet, event))
        {   action = (*zeta)[(UInt08)*state][i].action;
            *state = (*zeta)[(UInt08)*state][i].state;
            break;
        }
        i++;
    }
    if (action != NULL) action();  // Entsprechende Aktion ausführen (insofern vorhanden)
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IncomingStateMachine(): Verfahrensfunktion des Zustandsautomaten für eingehende     //
//  ----------------------  Nachrichten. Ausgangszustand ist ST_RECEIVE.                //
//                                                                                      //
//     Zustände:                                                                        //
//      · ST_RECEIVE    : Warten auf Empfang einer an uns adressierten Anfrage.         //
//      · ST_INBOX      : Anfrage liegt zur Abholung im Eingangsfach bereit.            //
//      · ST_IDLE       : Warten, bis Anfrage extern bearbeitet ist.                    //
//      · ST_OUTBOX     : Antwort liegt zur Übertragung im Ausgangsfach bereit.         //
//      · ST_TRANSMIT   : Antwort wird übertragen.                                      //
//                                                                                      //
//     Ereignisse:                                                                      //
//      · EV_RECEIVED   : An uns adressierte Anfrage ist eingetroffen.                  //
//      · EV_HANDLED    : Anfrage wurde abgeholt oder Antwort abgesetzt.                //
//      · EV_ADMISSION  : Freigabe zur Übertragung der Antwort ist erteilt.             //
//      · EV_TRANSMITTED: Übertragung des Antwort ist abgeschlossen.                    //
//      · EV_COLLISION  : Übertragungsfehler bzw. Kollision ist aufgetreten.            //
//      · EV_TIMEOUT    : Paketumlaufzeit ist abgelaufen.                               //
//      · EV_INVALID    : Zu sendende Antwort ist ungültig.                             //
//      · EV_BROADCAST  : Zu sendende Anfrage war ein Rundruf.                          //
//                                                                                      //
//                                                                                      //
//     Zustandsübergangsdiagramm:                                                       //
//                                                                                      //
//   EV_COLLISION                                                                       //
//   -> StartResponseBackoff() +-----------------+ EV_TRANSMITTED                       //
//   +-------------------------+   ST_TRANSMIT   +-------+                              //
//   |                         +-----------------+       |                              //
//   |                            /|\                    |                              //
//   |                             |                    \|/                             //
//   |                             |          +--------------------+                    //
//   |                             |     +--->|     ST_RECEIVE     |                    //
//   |                             |     |    +-----------------+--+                    //
//   |          +------------------+     |      /|\             | EV_RECEIVED:          //
//   |          |                        |       |              | -> AcquireRequest()   //
//   |          |                        |       |              |                       //
//   |          | EV_ADMISSION           |       |              |                       //
//   |          | -> TransmitResponse()  |       | EV_TIMEOUT  \|/                      //
//   |    +-----+---------+              |    +--+-----------------+                    //
//   +--->|   ST_OUTBOX   +--------------+    |      ST_INBOX      |                    //
//        +---------------+ EV_INVALID,       +---------+----------+                    //
//              /|\         EV_TIMEOUT,                 | EV_HANDLED                    //
//               |          EV_BROADCAST                |                               //
//               |                                      |                               //
//               |            +-------------+           |                               //
//               +------------+   ST_IDLE   |<----------+                               //
//                 EV_HANDLED +-------------+                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
IncomingStateMachine(void)
{
    static const ZetaItem zeta[][4] PROGMEM =
    {   { //- ST_RECEIVE --------------------------------------------------
            { EV_RECEIVED   , Action_AcquireRequest      , ST_INBOX    },
            { EV_NONE       , NULL                       , ST_RECEIVE  },
            { EV_NONE       , NULL                       , ST_RECEIVE  },
            { EV_NONE       , NULL                       , ST_RECEIVE  }
        },
        { //- ST_INBOX ----------------------------------------------------
            { EV_TIMEOUT    , NULL                       , ST_RECEIVE  },
            { EV_HANDLED    , NULL                       , ST_IDLE     },
            { EV_NONE       , NULL                       , ST_INBOX    },
            { EV_NONE       , NULL                       , ST_INBOX    }
        },
        { //- ST_IDLE -----------------------------------------------------
            { EV_HANDLED    , NULL                       , ST_OUTBOX   },
            { EV_NONE       , NULL                       , ST_IDLE     },
            { EV_NONE       , NULL                       , ST_IDLE     },
            { EV_NONE       , NULL                       , ST_IDLE     }
        },
        { //- ST_OUTBOX----------------------------------------------------
            { EV_INVALID    , NULL                       , ST_RECEIVE  },
            { EV_TIMEOUT    , NULL                       , ST_RECEIVE  },
            { EV_BROADCAST  , NULL                       , ST_RECEIVE  },
            { EV_ADMISSION  , Action_TransmitResponse    , ST_TRANSMIT }
        },
        { //- ST_TRANSMIT -------------------------------------------------
            { EV_COLLISION  , Action_StartResponseBackoff, ST_OUTBOX   },
            { EV_TRANSMITTED, NULL                       , ST_RECEIVE  },
            { EV_NONE       , NULL                       , ST_TRANSMIT },
            { EV_NONE       , NULL                       , ST_TRANSMIT }
        }
    };

    EventGenerator(&inEventSet, &inTimer, &inPause, &inPdu);
    StateTransition(&inState, &inEventSet, &zeta);
    ClearEvents(&inEventSet);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  OutgoingStateMachine(): Verfahrensfunktion des Zustandsautomaten für ausgehende     //
//  ----------------------  Nachrichten. Ausgangszustand ist ST_IDLE.                   //
//                                                                                      //
//     Zustände:                                                                        //
//      · ST_IDLE       : Keine Übertragung im Gange.                                   //
//      · ST_OUTBOX     : Anfrage liegt zur Übertragung im Ausgangsfach bereit.         //
//      · ST_TRANSMIT   : Anfrage wird übertragen.                                      //
//      · ST_RECEIVE    : Warten auf Empfang der passenden Antwort.                     //
//      · ST_INBOX      : Antwort liegt zur Abholung im Eingangsfach bereit.            //
//                                                                                      //
//     Ereignisse:                                                                      //
//      · EV_RECEIVED   : Eine zur Anfrage passende Anfrage ist eingetroffen.           //
//      · EV_HANDLED    : Anfrage wurde abgesetzt oder Antwort wurde abgeholt.          //
//      · EV_ADMISSION  : Freigabe zur Übertragung der Anfrage ist erteilt.             //
//      · EV_TRANSMITTED: Übertragung des Anfrage ist abgeschlossen.                    //
//      · EV_COLLISION  : Übertragungsfehler bzw. Kollision ist aufgetreten.            //
//      · EV_TIMEOUT    : Paketumlaufzeit ist abgelaufen.                               //
//      · EV_INVALID    : Zu sendende Anfrage ist ungültig.                             //
//      · EV_BROADCAST  : Zu sendende Anfrage ist ein Rundruf.                          //
//                                                                                      //
//                                                                                      //
//     Zustandsübergangsdiagramm:                                                       //
//                                                                                      //
//                                                                                      //
//    EV_COLLISION                                 EV_TRANSMITTED                       //
//    -> StartRequestBackoff() +-----------------+ -> StartTimer()                      //
//   +-------------------------+   ST_TRANSMIT   +----------+                           //
//   |                         +-----------------+          |                           //
//   |                                 /|\                  |                           //
//   |                                  |                  \|/                          //
//   |                                  |           +----------------+                  //
//   |                                  |           |   ST_RECEIVE   |                  //
//   |                                  |           +-------+--------+                  //
//   |                                  |                   | EV_RECEIVED               //
//   |            +---------------------+                   | -> AcquireResponse()      //
//   |            |                                         |                           //
//   |            |                                         | EV_TIMEOUT                //
//   |            |                                         | -> InvalidateResponse()   //
//   |            |                                         |                           //
//   |            |                                         | EV_BROADCAST              //
//   |            | EV_ADMISSION                            | -> VirtualResponse()      //
//   |            | -> TransmitRequest()                   \|/                          //
//   |    +-------+-------+                         +------------------+                //
//   +--->|   ST_OUTBOX   +------------------------>|     ST_INBOX     |                //
//        +---------------+ EV_INVALID,             +--------+---------+                //
//               /|\        EV_TIMEOUT                       | EV_HANDLED               //
//                |         -> InvalidateResponse()          |                          //
//                |                                          |                          //
//                | EV_HANDLED                               |                          //
//                | -> StartTimer() +-------------+          |                          //
//                +-----------------+   ST_IDLE   |<---------+                          //
//                                  +-------------+                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OutgoingStateMachine(void)
{
    static const ZetaItem zeta[][4] PROGMEM =
    {   { //- ST_RECEIVE -------------------------------------------------
            { EV_TIMEOUT    , Action_InvalidateResponse , ST_INBOX    },
            { EV_BROADCAST  , Action_VirtualResponse    , ST_INBOX    },
            { EV_RECEIVED   , Action_AcquireResponse    , ST_INBOX    },
            { EV_NONE       , NULL                      , ST_RECEIVE  }
        },
        { //- ST_INBOX ---------------------------------------------------
            { EV_HANDLED    , NULL                      , ST_IDLE     },
            { EV_NONE       , NULL                      , ST_INBOX    },
            { EV_NONE       , NULL                      , ST_INBOX    },
            { EV_NONE       , NULL                      , ST_INBOX    }
        },
        { //- ST_IDLE ----------------------------------------------------
            { EV_HANDLED    , Action_StartTimer         , ST_OUTBOX   },
            { EV_NONE       , NULL                      , ST_IDLE     },
            { EV_NONE       , NULL                      , ST_IDLE     },
            { EV_NONE       , NULL                      , ST_IDLE     }
        },
        { //- ST_OUTBOX --------------------------------------------------
            { EV_TIMEOUT    , Action_InvalidateResponse , ST_INBOX    },
            { EV_INVALID    , Action_InvalidateResponse , ST_INBOX    },
            { EV_ADMISSION  , Action_TransmitRequest    , ST_TRANSMIT },
            { EV_NONE       , NULL                      , ST_OUTBOX   }
        },
        { //- ST_TRANSMIT ------------------------------------------------
            { EV_COLLISION  , Action_StartRequestBackoff, ST_OUTBOX   },
            { EV_TRANSMITTED, Action_StartTimer         , ST_RECEIVE  },
            { EV_NONE       , NULL                      , ST_TRANSMIT },
            { EV_NONE       , NULL                      , ST_TRANSMIT }
        }
    };

    EventGenerator(&outEventSet, &outTimer, &outPause, &outPdu);
    StateTransition(&outState, &outEventSet, &zeta);
    ClearEvents(&outEventSet);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  SetupMessage(): Erzeugt aus dem angegebenen Datagramm eine Cy4-Nachricht. Ist das   //
//  --------------  übergebene Datagramm ungültig, ist die resultierende Cy4-Nachricht  //
//  ebenfalls ungültig.                                                                 //
//  SetupMessage() kommt in den Abholungsfunktionen CY4CORE_TakeRequest() und CY4CORE_- //
//  TakeResponse() zum Einsatz.                                                         //
//                                                                                      //
//  Eingabe: · pdu:     Eingabe-Datagramm.                                              //
//           · message: Referenz, die die generierte Cy4-Nachricht aufnimmt.            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupMessage(const PDU* pdu, Cy4Message* message)
{
    message->valid       = IsValid(pdu);
    message->access      = Access(pdu);
    message->address.net = pdu->byte[4];
    message->address.dev = pdu->byte[5];
    message->location    = ((UInt32)pdu->byte[6]<<16)
                           + (UInt32)((UInt16)pdu->byte[7]<< 8)
                           + pdu->byte[8];
    message->count       = pdu->byte[9];
    if (HasPayload(pdu))
    {   for (UInt08 i=0; i<message->count; i++) message->data[i] = pdu->byte[10+i];
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  SetupPdu(): Erzeugt aus der angegebenen Cy4-Nachricht ein Datagramm. Die Funktion   //
//  ----------  ermittelt selbständig, ob aus der Nachricht ein Anfrage-Datagramm für   //
//  die Ausgangs-Übertragungseinheit (outPdu) oder ein Antwort-Datagramm für die Ein-   //
//  gangs-Übertragungseinheit (inPdu) generiert werden muss.                            //
//  SetupPdu() kommt in den Ablieferungsfunktionen CY4CORE_PostRequest() und CY4CORE_-  //
//  PostResponse() zum Einsatz.                                                         //
//                                                                                      //
//  Eingabe: · message:  Cy4-Nachricht.                                                 //
//           · pdu:      Referenz, die das generierte Datagramm aufnimmt.               //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupPdu(const Cy4Message* message, PDU* pdu)
{
    Invalidate(pdu);                        // Datagramm vorläufig für ungültig erklären
    if (!message->valid) return;            // Ungültige Nachrichten nicht behandeln
    bool createRequest = pdu == &outPdu;    // 'true' = Datagramm ist eine Anfrage
    // DTF bestimmen
    UInt08  dtf;                                            //  DTF | READ | WRITE
    if (message->access == CY4MESSAGE_WRITE) dtf = 0x0;     //  ----+------+------
    else                                     dtf = 0x1;     //  REQ | C2h  | C3h
    if (createRequest) dtf = ~dtf;                          //  RSP | C1h  | C0h
    dtf = (PDU_DIB<<2) | (0x03 & dtf);
    // Adressat festlegen (Anfrage) bzw. prüfen (Antwort)
    Cy4Address destinationAddress = message->address;
    if (!createRequest)
    {   if (destinationAddress.net != pdu->byte[4] ||
            destinationAddress.dev != pdu->byte[5])
        {   // Antwort an falschen Adressaten
            return;
        }
        if (IsBroadcast(pdu))
        {   // Antwort auf einen Rundruf
            destinationAddress.net = 0;
            destinationAddress.dev = 0;
        }
    }
    // Absendeadresse festlegen
    Cy4Address sourceAddress;
    if (createRequest) sourceAddress = selfAddress;
    else                sourceAddress = CY4ADDRESS_Create(pdu->byte[2], pdu->byte[3]);
    // Datagramm zusammenstellen
    pdu->byte[0] = PDU_DPS;
    pdu->byte[1] = dtf;
    pdu->byte[2] = destinationAddress.net;
    pdu->byte[3] = destinationAddress.dev;
    pdu->byte[4] = sourceAddress.net;
    pdu->byte[5] = sourceAddress.dev;
    pdu->byte[6] = (UInt08)(message->location >> 16);
    pdu->byte[7] = (UInt08)(message->location >>  8);
    pdu->byte[8] = (UInt08)(message->location >>  0);
    pdu->byte[9] = message->count;
    if (pdu->byte[9] > CY4_PAYLOAD_CAPACITY) pdu->byte[9] = 0;
    UInt08 crc = 0xFF;
    UInt08 i;
    for (i=0; i<10; i++) UpdateCRC(pdu->byte[i], &crc);
    if (HasPayload(pdu))
    {   for (i=0; i<pdu->byte[9]; i++)
        {   pdu->byte[10+i] = message->data[i];
            UpdateCRC(pdu->byte[10+i], &crc);
        }
        i += 10;
    }
    pdu->byte[i] = crc;
    if (IsBroadcast(pdu) && !IsExxxxxWriteRequest(pdu))
    {   // Datagramm ist entweder keine gültige Rundruf-Anfrage oder eine Rundruf-Antwort
        Invalidate(pdu);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_SetAddress(): Siehe CY4NET_SetAddress().                                    //
//  --------------------                                                                //
//                                                                                      //
//  Eingabe: · address: Cy4NET-Adresse des eigenen Gerätes.                             //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
CY4CORE_SetAddress(Cy4Address address)
{
    selfAddress = address;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_Address(): Siehe CY4NET_Address().                                          //
//  -----------------                                                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Cy4NET-Adresse des eigenen Gerätes.                                        //
//////////////////////////////////////////////////////////////////////////////////////////
Cy4Address
CY4CORE_Address(void)
{
    return selfAddress;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_TakeRequest(): Empfängt die Übertragungseinheit eine Anfrage, dient diese   //
//  ---------------------  Funktion dazu die Anfrage entgegenzunehmen. Eine einge-      //
//  gangene Anfrage muss angenommen, bearbeitet und die Antwort mittels CY4CORE_Post-   //
//  Response() wieder zurückgeschickt werden.                                           //
//                                                                                      //
//  Eingabe: · message: Referenz, die eine bereitstehende Anfrage aufnimmt. Steht zum   //
//                      Zeitpunkt des Aufrufs keine Anfrage zur Abholung bereit, ist    //
//                      die in <message> abgelegte Nachricht ungültig.                  //
//  Ausgabe: 'true', wenn eine Anfrage zur Abholung bereitstand, wenn also <message>    //
//           eine gültige Anfrage erhalten hat.                                         //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4CORE_TakeRequest(Cy4Message* message)
{
    CY4MESSAGE_Invalidate(message);          // Nachricht vorerst ungültig ...
    if (inState != ST_INBOX) return false;   // Wartet eine Anfrage im Eingangsfach?
    SetupMessage(&inPdu, message);           // Anfrage-Nachricht aufstellen
    AddEvent(&inEventSet, EV_HANDLED);       // Ereignis: Anfrage abgeholt
    IncomingStateMachine();                  // Eingangs-Übertragungseinheit aktualisieren
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_PostResponse(): Initiiert das Senden einer Antwort auf eine zuvor mittels   //
//  ----------------------  CY4CORE_TakeRequest() entgegengenommene Anfrage. War die    //
//  Anfrage falsch oder konnte, aus welchem Grund auch immer, nicht bearbeitet werden,  //
//  ist die übergebende Antwort-Nachricht vorher für ungültig zu erklären (CY4MESSAGE_- //
//  Invalidate).                                                                        //
//                                                                                      //
//  Eingabe: · message: Die zu sendenden Antwort-Nachricht.                             //
//  Ausgabe: 'true', wenn die Nachricht zur Absendung entgegengenommen wurde. 'false',  //
//           wenn die Übertragungseinheit noch mit dem Versenden einer anderen Antwort  //
//           beschäftigt ist.                                                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4CORE_PostResponse(const Cy4Message* message)
{
    if (inState != ST_IDLE) return false;   // Eingangs-Übertragungseinheit bereit?
    SetupPdu(message, &inPdu);              // Antwort-Datagramm aufstellen
    AddEvent(&inEventSet, EV_HANDLED);      // Ereignis: Antwort abgeliefert
    IncomingStateMachine();                 // Eingangs-Übertragungseinheit aktualisieren
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_PostRequest(): Initiiert das Senden einer Anfrage. Die Funktion kehrt un-   //
//  ---------------------  mittelbar zum Aufrufer zurück. Die Übertragungseinheit ist   //
//  in der Lage stets nur eine einzelne Anfrage-Antwort-Sequenz zu handhaben. Folge     //
//  ist, dass weitere Anfragen solange abgewiesen werden, wie eine laufende Über-       //
//  tragungssequenz noch nicht abgeschlossen ist. Den Abschluss einer Übertragungs-     //
//  sequenz bildet das Abholen der Antwort mittels CY4CORE_TakeResponse().              //
//                                                                                      //
//  Eingabe: · message: Die zu sendende Anfrage-Nachricht.                              //
//  Ausgabe: 'true', wenn die Nachricht zur Absendung entgegengenommen wurde.           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4CORE_PostRequest(const Cy4Message* message)
{
    if (outState != ST_IDLE) return false;  // Ausgangs-Übertragungseinheit bereit?
    SetupPdu(message, &outPdu);             // Anfrage-Datagramm aufstellen
    AddEvent(&outEventSet, EV_HANDLED);     // Ereignis: Anfrage abgeliefert
    OutgoingStateMachine();                 // Ausgangs-Übertragungseinheit aktualisieren
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_TakeResponse(): Jede von CY4CORE_PostRequest() entgegengenommene Anfrage    //
//  ----------------------  wird beantwortet. Steht bei Aufruf dieser Funktion eine     //
//  Antwort zur Abholung bereit, wird sie entgegengenommen und in <message> abgelegt.   //
//  War die Kommunikation erfolgreich, ist die Antwort-Nachricht gültig und führt die   //
//  gewünschten Daten mit sich. Andernfalls erhält man als Antwort eine ungültige Kopie //
//  der Anfrage.                                                                        //
//  Solange eine bereitstehenden Antwort nicht abgeholt ist, wird CY4CORE_PostRequest() //
//  die Annahme neuer Anfragen verweigern.                                              //
//                                                                                      //
//  Eingabe: · message: Referenz, die eine eventuell bereitstehende Antwort aufnimmt.   //
//                      Steht zum Zeitpunkt des Aufrufs keine Antwort zur Abholung      //
//                      bereit, ist die in <message> abgelegte Nachricht ungültig.      //
//  Ausgabe: 'true', wenn in <message> eine abgeholte Antwort abgelegt wurde.           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4CORE_TakeResponse(Cy4Message* message)
{
    CY4MESSAGE_Invalidate(message);          // Nachricht vorerst ungültig ...
    if (outState != ST_INBOX) return false;  // Wartet eine Antwort im Eingangsfach?
    SetupMessage(&outPdu, message);          // Antwort-Nachricht aufstellen
    AddEvent(&outEventSet, EV_HANDLED);      // Ereignis: Antwort abgeholt
    OutgoingStateMachine();                  // Ausgangs-Übertragungseinheit aktualisieren
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4CORE_IsIdle(): Prüft, ob die Übertragungseinheit zur Absetzung einer Anfrage     //
//  ----------------  mittels CY4CORE_PostRequest() bereit ist. Das ist dann der Fall,  //
//  wenn aktuell keine Anfrage gesendet wird und keine eingegangene Antwort noch auf    //
//  Abholung wartet.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn die Übertragungseinheit bereit ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4CORE_IsIdle(void)
{
    return outState == ST_IDLE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
CY4CORE_Startup(void)
{
    inState            = ST_RECEIVE;
    outState           = ST_IDLE;
    newPduReceived     = false;
    Cy4Address address = CY4ADDRESS_INVALID;
    selfAddress        = address;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
CY4CORE_Proceed(void)
{
    ProceedReceiving();
    IncomingStateMachine();
    OutgoingStateMachine();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
CY4CORE_Shutdown(void)
{
    while (inState == ST_OUTBOX || inState == ST_TRANSMIT) IncomingStateMachine();
}

//////////////////////////////////////////////////////////////////////////////////////////
