﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//      The Cy4NET System                      Version 19.10                            //
//      __________________________             _________________________________        //
//                                                                                      //
//      Decentralised Realtime                 Copyright (c) 2012 Reimar Grasbon        //
//      Peer-to-peer Communication             (www.cy4net.org)                         //
//      Protocol                                                                        //
//      _________________________              _________________________________        //
//                                                                                      //
//      [Base Unit]                                                                     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "ARMChip.h"
#include "Cy4Net.h"
#include "Cy4Core.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct                              // Element der Versende-Schlange:
{   Cy4MessageSetup    messageSetup;        // · Cy4-Nachrichten-Ausstattungsfunktion
    Cy4ResponseHandler responseHandler;     // · Cy4-Dienst Antwortbehandler
} PostingItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static const Cy4ServiceItem* servicesTable      = NULL;
static UInt08                servicesTableSize  = 0;
static UInt08                postingQueueHead   = 0;
static UInt08                postingQueueLength = 0;
static PostingItem           postingQueueBuffer[CFG_CY4_POSTING_QUEUE_CAPACITY];

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  EnqueuePosting(): Stellt <item> als neues letztes Element in die Versende-Schlange. //
//  ----------------  Ist die Schlange voll, bleibt die Funktion ohne Wirkung. Die      //
//  Versende-Schlange hat Platz für <CFG_CY4_POSTING_QUEUE_CAPACITY> Elemente.          //
//                                                                                      //
//  Eingabe: · item: Das anzuhängende Versende-Element.                                 //
//  Ausgabe: 'true', wenn noch Platz vorhanden war und das Element aufgenommen wurde.   //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
EnqueuePosting(const PostingItem* item)
{
    if (postingQueueLength == CARDINALITY(postingQueueBuffer)) return false;
    UInt08 tail = postingQueueHead + postingQueueLength;
    if (tail >= CARDINALITY(postingQueueBuffer)) tail -= CARDINALITY(postingQueueBuffer);
    postingQueueBuffer[tail] = *item;
    postingQueueLength++;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DequeuePosting(): Entfernt das vorderste Element aus der Versende-Schlange. Ist die //
//  ----------------  Schlange leer, bleibt die Funktion ohne Wirkung.                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
DequeuePosting(void)
{
    if (postingQueueLength == 0) return;
    postingQueueHead++;
    if (postingQueueHead == CARDINALITY(postingQueueBuffer)) postingQueueHead = 0;
    postingQueueLength--;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  FirstPosting(): Liefert einen Zeiger auf das vorderste Element der Versende-        //
//  --------------  Schlange zurück. Ist die Schlange leer, ist der zurückgelieferte    //
//  Zeiger NULL.                                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Zeiger auf das vorderste Element der Versende-Schlange oder NULL.          //
//////////////////////////////////////////////////////////////////////////////////////////
static PostingItem*
FirstPosting(void)
{
    if (postingQueueLength == 0) return NULL;
    return &postingQueueBuffer[postingQueueHead];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ContainsPosting(): Prüft, ob <item> in der Versende-Schlange enthalten ist. Das ist //
//  -----------------  dann der Fall, wenn beide Funktionszeiger in <item> mit den      //
//  Zeigern (mindestens) eines Eintrages der Schlage übereinstimmen.                    //
//                                                                                      //
//  Eingabe: · item: Das zu prüfende Versende-Element.                                  //
//  Ausgabe: 'true', wenn <item> in der Schlange vorkommt                               //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
ContainsPosting(const PostingItem* item)
{
    UInt08 position = postingQueueHead;
    for (UInt08 i=0; i<postingQueueLength; i++)
    {   const PostingItem* tmpItem = &postingQueueBuffer[position];
        if (tmpItem->messageSetup    == item->messageSetup &&
            tmpItem->responseHandler == item->responseHandler) return true;
        position++;
        if (position == CARDINALITY(postingQueueBuffer)) position = 0;
    }
    return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  HandleRequest(): Sucht für die Anfrage in <message> den passenden Dienst aus der    //
//  ---------------  zuvor mittels CY4NET_AssignServices() festgelegten Tabelle von     //
//  Diensten aus. Findet sich ein passender Dienst, ruft HandleRequest() den zuge-      //
//  hörigen Dienstbearbeiter auf und übergibt ihm die Anfrage in <message>. Der Be-     //
//  arbeiter bearbeitet die Anfrage und platziert seine Antwort anschliessend wieder in //
//  <message>, die dann von HandleRequest() an die aufrufende Umgebung zurückgegeben    //
//  wird. Konnte der Dienstbearbeiter die Anfrage, aus welchen Gründen auch immer,      //
//  nicht bearbeiten oder hat seine Antwort verworfen, muss der Dienstbearbeiter die    //
//  Antwortnachricht in <message> für Ungültig erlären. Ebenfalls ungültig ist die      //
//  Nachricht in <message>, wenn keinen passender Dienst in der Liste gefunden wurde.   //
//                                                                                      //
//  Eingabe: · message: Nachricht, zu der der passende Dienste gesucht und dessen       //
//                      Bearbeiter aufgerufen werden soll. In <message> befindet        //
//                      anschließend die Antwort des Dienstbearbeiters.                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
HandleRequest(Cy4Message* message)
{
    if (!message->valid) return;
    bool valid = false;
    if (servicesTableSize != 0)
    {   UInt32 messageLocation = message->location;
        UInt32 serviceBase = servicesTable[0].service.base;
        if (messageLocation >= serviceBase)
        {   // (1) Passenden Dienst suchen
            SInt08 a = 0;
            SInt08 b = (SInt08)servicesTableSize-1;
            SInt08 m;
            do
            {   m = a + (b-a)/2;
                serviceBase = servicesTable[m].service.base;
                if      (messageLocation > serviceBase) a = m+1;
                else if (messageLocation < serviceBase) b = m-1;
                else                                    b = m;
            } while (a <= b && b != m);
            Cy4ServiceItem  item = servicesTable[b];
            // (2) Zugriffsberechtigung prüfen
            Cy4MessageAccess messageAccess = message->access;
            Cy4ServiceAccess serviceAccess = item.service.access;
            if ((messageAccess == CY4MESSAGE_READ  && serviceAccess != CY4SERVICE_WRITE_ONLY) ||
                (messageAccess == CY4MESSAGE_WRITE && serviceAccess != CY4SERVICE_READ_ONLY ))
            {   // (3) Zugriffsgrenzen prüfen
                UInt08 messageCount = message->count;
                UInt16 serviceSize  = item.service.size;
                serviceBase = item.service.base;
                if (serviceBase + serviceSize >= messageLocation + messageCount)
                {   // (4) Dienst-Anfragebearbeiter aufrufen
                    if (item.handler != NULL)
                    {   item.handler(&item.service, message);
                        valid = message->valid;
                    }
                }
            }
        }
    }
    if (!valid) CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProcessIncoming(void)
{
    Cy4Message  message;

    if (CY4CORE_TakeRequest(&message))
    {   // Anfrage erhalten, Anfragebearbeiter aufrufen, Antwort zurücksenden
        HandleRequest(&message);
        CY4CORE_PostResponse(&message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProcessOutgoing(void)
{
    PostingItem* item = FirstPosting();
    if (item == NULL) return;
    if (item->messageSetup != NULL)
    {   // Anfrage
        if (CY4CORE_IsIdle())
        {   Cy4Message message = CY4MESSAGE_INVALID;
            (item->messageSetup)(&message);
            if (CY4MESSAGE_IsValid(&message))
            {   item->messageSetup = NULL;  // Markierung für: Sendung unterwegs
                if (message.address.net == CY4CORE_Address().net &&
                    message.address.dev == CY4CORE_Address().dev)
                {   // Lokale Nachricht unmittelbar beantworten
                    HandleRequest(&message);
                    Cy4ResponseHandler responseHandler = item->responseHandler;
                    DequeuePosting();
                    if (responseHandler != NULL) responseHandler(&message);
                }
                else
                {   CY4CORE_PostRequest(&message);
                }
            }
            else
            {   // Erzeugte Nachricht ist ungültig, Posting verwerfen
                DequeuePosting();
            }
        }
    }
    else
    {   // Antwort
        Cy4Message message;
        if (CY4CORE_TakeResponse(&message))
        {   Cy4ResponseHandler responseHandler = item->responseHandler;
            DequeuePosting();
            if (responseHandler != NULL) responseHandler(&message);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_AssignServices(): Übergibt dem Cy4NET-System die Dienste, die das jeweilige  //
//  -----------------------  Gerät bereitstellt. Dazu erwartet die Funktion in <table>  //
//  die Adresse eines Feldes mit Einträgen vom Typ 'Cy4ServiceItem' im Programm-        //
//  speicher. Ein Cy4ServiceItem besteht aus der Dienstsignatur, also Dienstort, Größe  //
//  und Zugriffsart und einer Bearbeiter-Rückruffunktion. Die Bearbeiterfunktion wird   //
//  aufgerufen sobald bei der Übertragungseinheit eine passende Dienstanfrage einge-    //
//  gangen ist.                                                                         //
//                                                                                      //
//    Beispiel:                                                                         //
//                                                                                      //
//    static const Cy4ServiceItem  servicesTable[] PROGMEM =                            //
//    {   //|                   service                    |        handler       |     //
//        //+--------------------------------------.....---+----------------------+     //
//        { CY4SERVICE(0xFFFF00, 61, CY4SERVICE_READ_WRITE), OnDeviceLabel        },    //
//        { CY4SERVICE(0xFFFF40,  1, CY4SERVICE_WRITE_ONLY), OnFactorySettings    },    //
//        { CY4SERVICE(0xFFFFF0,  7, CY4SERVICE_READ_ONLY ), OnCallTypePlate      },    //
//        { CY4SERVICE(0xFFFFF8,  1, CY4SERVICE_WRITE_ONLY), OnEnterApplication   },    //
//        { CY4SERVICE(0xFFFFFA,  1, CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader },    //
//        { CY4SERVICE(0xFFFFFC,  2, CY4SERVICE_WRITE_ONLY), OnChangeCy4Address   }     //
//    };                                                                                //
//                                                                                      //
//  Es ist darauf zu achten, dass die Einträge der Tabelle nach Dienstadressen auf-     //
//  steigend sortiert aufgelistet sind!                                                 //
//  Den Abschluss der Liste müssen immer die vier Pflichtdienste CallTypePlate, Enter-  //
//  Application, EnterProgramLoader und ChangeCy4Address bilden (siehe Beispiel).       //
//  Die Dienste-Tabelle kann je nach Bedarf, auch im laufenden Betrieb, getauscht       //
//  werden und darf bis zu 127 Einträge haben, Pflichtdienste eingerechnet.             //
//                                                                                      //
//  Eingabe: · table: Adresse der Dienste-Tabelle (im Programmspeicher).                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
CY4NET_AssignServices(const Cy4ServiceItem* table)
{
    UInt08 i=0;
    if (table != NULL)
    {   while (table[i].service.base != 0xFFFFFC) i++;
        i++;
    }
    servicesTable     = table;
    servicesTableSize = i;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_SetAddress(): Setzt die eigene Cy4NET-Geräteadresse. Die aufrufende Umgebung //
//  -------------------  muss dabei sicherstellen, daß <address> gültig (nicht 255.255) //
//  und keine Rundrufadresse (dev und net ungleich null) ist.                           //
//                                                                                      //
//  Eingabe: · address: Cy4NET-Adresse des eigenen Gerätes.                             //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
CY4NET_SetAddress(Cy4Address address)
{
    CY4CORE_SetAddress(address);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_Address(): Liefert die aktuelle eingestellte Cy4NET-Geräteadresse.           //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Cy4NET-Adresse des eigenen Gerätes.                                        //
//////////////////////////////////////////////////////////////////////////////////////////
Cy4Address
CY4NET_Address(void)
{
    return CY4CORE_Address();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_QueueLength(): Liefert die Anzahl auf Aussendung wartender Sendungen in der  //
//  --------------------  Schlange. Solange die Funktion einen Wert kleiner 'CFG_CY4_-  //
//  POSTING_QUEUE_CAPACITY' zurückliefert, ist die Versende-Schlange nicht voll, d.h.   //
//  CY4NET_Post() wird die Annahme weiterer Anfragen akzeptieren.                       //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Anzahl wartender Sendungen: 0 .. CFG_CY4_POSTING_QUEUE_CAPACITY.           //
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
CY4NET_QueueLength(void)
{
    return postingQueueLength;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_Post(): Nimmt eine Anfrage entgegen und stellt sie als neues letztes Ele-    //
//  -------------  ment in die Schlange auszugehender Sendungen.                        //
//  Eine  Anfrage ist bestimmt durch die Funktionszeiger <messageSetup> und <response-  //
//  Handler>. Ist die Anfrage an der Reihe, wird vor ihrer Verschickung die durch <mes- //
//  sageSetup> referenzierte Funktion ausgeführt, die das bereitgestellte Cy4Message-   //
//  Objekt mit der gewünschten Anfrage ausstattet. Lässt die Ausstattungsfunktion das   //
//  mitgegebene Cy4Message-Objekt unangetastet bzw. ist die erzeugte Cy4Message un-     //
//  gültig, wird die Anfrage ignoriert und aus der Schlange entfernt.                   //
//  Nach Erhalt der Antwort wird der in <responseHandler> zu Verfügung gestellte Ant-   //
//  wortbehandler aufgerufen, dem als Parameter die erhaltene Antwort-Nachricht mitge-  //
//  liefert wird. Ist diese Nachricht gültig, war die Kommunikation erfolgreich. An-    //
//  dernfalls war die Anfrage fehlerhaft oder die Übertragung ist gescheitert. Die      //
//  übergebene Cy4Message ist in dem Fall eine als ungültig gekennzeichnete Kopie der   //
//  Anfrage.                                                                            //
//  Im Gegensatz zur Ausstattungsfunktion in <messageSetup> ist die Angabe eines Ant-   //
//  wortbehandlers optional, d.h. <responseHandler> darf NULL sein.                     //
//                                                                                      //
//  Hinweis: Rundrufe werden nie beantwortet. Trotz dessen wird aber auch für Rundrufe  //
//           der Antwort-Behandler aufgerufen. Als Antwort-Nachricht erhält der Be-     //
//           handler in dem Fall eine Kopie der Anfrage, deren Gültigkeit darüber in-   //
//           formiert, ob das System den Rundruf hat absetzen können oder nicht. Das    //
//           ist insofern kein Problem, da Rundrufe stets Schreibaufforderungen sind,   //
//           deren Antworten keine Nutzlast befördern. Ob der Rundruf letztlich alle    //
//           Zielknoten erreicht hat, bleibt ungewiss.                                  //
//                                                                                      //
//  Die Aufnahmekapazität der Schlange ist durch <CFG_CY4_POSTING_QUEUE_CAPACITY> fest- //
//  gelegt. Man beachte, dass sich bei auf Aufreihung von Anfragen deren Antwortzeiten  //
//  erhöhen können.                                                                     //
//                                                                                      //
//  Eingabe: · messageSetup:    Cy4-Nachrichten-Ausstattungsfunktion.                   //
//           · responseHandler: Cy4-Dienst Antwortbehandler oder NULL.                  //
//  Ausgabe: 'true', wenn die Anfrage zur Absendung entgegengenommen wurde, 'false',    //
//           wenn die Schlange voll ist und daher die Annahme der Anfrage verweigert    //
//           wurde (dazu s.a. CY4NET_QueueLength()).                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4NET_Post(Cy4MessageSetup messageSetup, Cy4ResponseHandler responseHandler)
{
    PostingItem item = { messageSetup, responseHandler };
    return EnqueuePosting(&item);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CY4NET_IsPending(): Prüft, ob die durch <messageSetup> und <responseHandler>        //
//  ------------------  bestimmte Anfrage bereits in der Schlange auf Absendung wartet. //
//  Die Information darüber, ob eine Anfrage gleichen Typs bereits auf Versendung       //
//  wartet ermöglicht es, redundante Anfragen zu vermeiden, Buskommunikation zu ver-    //
//  ringern und einem Volllaufen der Sendeschlange entgegenzuwirken.                    //
//                                                                                      //
//  Eingabe: · messageSetup:    Cy4-Nachrichten-Ausstattungsfunktion.                   //
//           · responseHandler: Cy4-Dienst Antwortbehandler oder NULL.                  //
//  Ausgabe: 'true', wenn bereits eine entsprechende Anfrage in der Sendeschlange       //
//           wartet.                                                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
CY4NET_IsPending(Cy4MessageSetup messageSetup, Cy4ResponseHandler responseHandler)
{
    PostingItem item = { messageSetup, responseHandler };
    return ContainsPosting(&item);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
CY4NET_Startup(void)
{
    CY4CORE_Startup();
    servicesTable      = NULL;
    servicesTableSize  = 0;
    postingQueueHead   = 0;
    postingQueueLength = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
CY4NET_Proceed(void)
{
    CY4CORE_Proceed();
    ProcessIncoming();
    ProcessOutgoing();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
CY4NET_Shutdown(void)
{
    while (!CY4CORE_IsIdle())
    {   ProcessOutgoing();
        CY4CORE_Proceed();
    }
    CY4CORE_Shutdown();
}

//////////////////////////////////////////////////////////////////////////////////////////
