﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Datenaufbewahrung                                                                   //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2022 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "ARMChip.h"
#include "Flash.h"
#include "Timer.h"
#include "Storing.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct  // Wertespeicherstruktur
{   Storing		storing;
    UInt16		crc;
} StoringContainer __attribute__ ((aligned));

//////////////////////////////////////////////////////////////////////////////////////////
//  Konstanten  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
const Storing factoryDef PROGMEM = // Werksvoreinstellungen
{   .STO_DeviceLabel                = "[Cy4NET F303 Nucleo Demo]",
	.STO_LedSignal                  = { LEDSIGNAL_HEARTBEAT, LEDSIGNAL_BUS_ACTIVITY, LEDSIGNAL_PUSH_BUTTON },
	.STO_PressedActionConfiguration = { false, PUSHMESSAGE_INVALID, TEXTFEED_EMPTY }
};

extern UInt08 STORING_SECTION_START;		// Definiert im ...
extern UInt08 CY4ADR_SECTION_START[8];		// ... Linker-Command-File

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool             modified = false;
static StoringContainer storingContainer;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
// 	CalcCRC(): Ermittelt die Prüfsumme des angegebenen Wertespeichers. 					//
//  ---------  																			//
//																						//
//  Eingabe: · storing: Addresse des Wertespeichers.   							        //
//  Ausgabe: CRC-16-Prüfsumme.															//
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
CalcCRC(const Storing* storing)
{
    UInt16 crc = 0xFFFF;
    for (UInt16 i=0; i<sizeof(storingContainer.storing); i++)
    {   UInt08 byte = ((UInt08*)storing)[i];
        crc ^= byte;
        for (UInt08 i=8; i>0; i--) crc = (crc >> 1) ^ ((crc&1) ? 0x8408 : 0);
    }
    if (crc == 0xFFFF) crc = 0; 	// Prüfsummenwert 0xFFFF umgehen
    return crc;
}

//////////////////////////////////////////////////////////////////////////////////////////
// 	IsBlank(): Prüft, ob der Wertespeicher-Container unbeschrieben (komplett 0xFF) ist. //
//  ---------  																			//
//																						//
//  Eingabe: · container: Addresse des zu prüfenden Wertespeicher-Containers.   	    //
//  Ausgabe: true, Speicherbereich ist leer.											//
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsBlank(const StoringContainer* container)
{
	for (UInt16 i=0; i<sizeof(StoringContainer); i++)
	{   UInt08 byte = ((UInt08*)container)[i];
		if (byte != 0xFF) return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
// 	ClearAll(): Löscht alle für die Datenaufbewahrung vorgesehenen Speicherseiten.		//
//  ----------  Da die Cy4-Adresssektion Bestandteil der Datenaufbewahrungssektion ist, //
//  wird die Addresse gerettet und dach dem Löschvorgang wieder restauriert.		    //
//																						//
//  Eingabe: -																			//
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ClearAll(void)
{
	Cy4Address address = STORING_LoadCy4SelfAddress();
	PageNr pageNr = FLASH_PageNumberOf((const void*)&STORING_SECTION_START);
	while (pageNr.number < FLASH_PAGE_COUNT)
	{	FLASH_ErasePage(pageNr);
		pageNr.number++;
	}
	STORING_SaveCy4SelfAddress(address);
}

//////////////////////////////////////////////////////////////////////////////////////////
// 	RecentContainer(): Ermittelt den zuletzt geschriebenen Wertespeicher-Container.	    //
//	-----------------  										                            //
//																						//
//  Eingabe: -																			//
//  Ausgabe: Adresse des Containers, NULL wenn keiner existiert.			            //
//////////////////////////////////////////////////////////////////////////////////////////
static const StoringContainer*
RecentContainer(void)
{
	const StoringContainer* recent = NULL;
	const StoringContainer* next   = (StoringContainer*)&STORING_SECTION_START;
	const StoringContainer* stop   = (StoringContainer*)((UInt32)&CY4ADR_SECTION_START -
	                                                    sizeof(StoringContainer));
	while (CalcCRC(&next->storing) == next->crc)
	{	recent = next;
		next++;
		if (next > stop) return recent;
	}
	if (!IsBlank(next)) recent = NULL;
	return recent;
}

//////////////////////////////////////////////////////////////////////////////////////////
// 	NextContainerPlace(): Sucht vorwärts nach dem nächsten freien Platz für die Ablage 	//
//	--------------------  eines weiteren Wertespeicher-Containers.			            //
//																						//
//  Eingabe: -																			//
//  Ausgabe: Adresse des Platzes.			        								    //
//////////////////////////////////////////////////////////////////////////////////////////
static StoringContainer*
NextContainerPlace(void)
{
	StoringContainer*       recent = (StoringContainer*)RecentContainer();
	StoringContainer*       next   = recent+1;
	const StoringContainer* last   = (StoringContainer*)((UInt32)&CY4ADR_SECTION_START -
		                                                sizeof(StoringContainer));
	if (recent == NULL || next > last)
	{	ClearAll();
		next = (StoringContainer*)&STORING_SECTION_START;
	}
	return next;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ReadStoring(void)
{
	const StoringContainer* recent = RecentContainer();
	if (recent == NULL) STORING_RestoreDefaults();
	else    			storingContainer = *recent;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
WriteStoring(void)
{
	// (1) CRC des Wertespeichers ergänzen
	storingContainer.crc = CalcCRC(&storingContainer.storing);
	// (2) Startadresse der nächsten freien Container-Platzes holen
	StoringContainer* address = NextContainerPlace();
	// (3) Entsprechende Seite aus dem Flash nach <data> laden
	FLASH_DirectWrite(&storingContainer, sizeof(StoringContainer), address);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_LoadBlock(const void* memAddress, void* ramAddress, UInt16 count)
{
	UInt32  storingBase = (UInt32)(&(storingContainer.storing));
	UInt08* source      = (UInt08*)(storingBase + (UInt32)memAddress);
	for (int i=0; i<count; i++)
	{	((UInt08*)ramAddress)[i] = source[i];
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_SaveBlock(const void* ramAddress, void* memAddress, UInt16 count)
{
	UInt32  storingBase = (UInt32)(&(storingContainer.storing));
	UInt08* destination = (UInt08*)(storingBase + (UInt32)memAddress);
	for (int i=0; i<count; i++)
	{	UInt08 byte = ((UInt08*)ramAddress)[i];
		if (destination[i] != byte)
		{	destination[i]  = byte;
			modified = true;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_RestoreDefaults(void)
{
	storingContainer.storing = factoryDef;
	if (!IsBlank((const StoringContainer*)&STORING_SECTION_START)) ClearAll();
}

//////////////////////////////////////////////////////////////////////////////////////////
Cy4Address
STORING_LoadCy4SelfAddress(void)
{
	const UInt08* baseAddress = (const UInt08*)(&CY4ADR_SECTION_START);
    UInt08 invAddress[] = { ~baseAddress[2], ~baseAddress[3] };
    Cy4Address address;
    if (baseAddress[0] == invAddress[0] && baseAddress[1] == invAddress[1])
    {   address.net = baseAddress[0];
        address.dev = baseAddress[1];
    }
    else
    {   address.net = CFG_DEFAULT_CY4_ADDRESS_NET;
        address.dev = CFG_DEFAULT_CY4_ADDRESS_DEV;
        STORING_SaveCy4SelfAddress(address);
    }
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_SaveCy4SelfAddress(Cy4Address address)
{
	PageNr pageNr = FLASH_PageNumberOf(&CY4ADR_SECTION_START);
	UInt08 pageData[FLASH_PAGE_SIZE];
	FLASH_ReadPage(pageNr, pageData);
	UInt32 index = ((UInt32)(&CY4ADR_SECTION_START) & (FLASH_PAGE_SIZE-1));
	pageData[index+0] =  address.net;
	pageData[index+1] =  address.dev;
	pageData[index+2] = ~address.net;
	pageData[index+3] = ~address.dev;
    FLASH_WritePage(pageData, pageNr);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Startup(void)
{
	ReadStoring();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Proceed(void)
{
	static Timer timer   = TIMER_MILLISECONDS(CFG_FLASH_STORING_DEFERMENT);
	static bool  pending = false;

	if (modified)
	{	modified = false;
		TIMER_Restart(&timer);
		pending = true;
	}
	if (pending && TIMER_HasExpired(&timer))
	{	pending = false;
		WriteStoring();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Shutdown(void)
{
	if (modified) WriteStoring();
}

//////////////////////////////////////////////////////////////////////////////////////////
